ACTION_IDENTIFIER = "_ACTION"
KEY_IDENTIFIER = "_KEY"

# - Action Types:
LOAD_URL_ACTION = "load_url"
CLICK_ACTION = "click"
CLICK_ELEMENT_WITH_OFFSET_ACTION = "click_element_with_offset"
HOVER_ACTION = "hover"
SCROLL_AN_ELEMENT_ACTION = "scroll_an_element"
REFRESH_ACTION = "refresh"
ENTER_TEXT_ACTION = "enter_text"
SEND_SPECIAL_KEY_ACTION = "send_special_key"
SLEEP_ACTION = "sleep"
WAIT_FOR_ELEMENT_ACTION = "wait_for_element"
SCROLL_WINDOW_TO_POSITION_ACTION = "scroll_window_to_position"
SCROLL_WINDOW_TO_ELEMENT_ACTION = "scroll_window_to_element"
FOR_EACH_ACTION = "for_each"
SHOW_ELEMENT_ACTION = "show_element"
HIDE_ELEMENT_ACTION = "hide_element"
SWITCH_WINDOW_HANDLE_ACTION = "switch_window_handle"
CLOSE_WINDOW_ACTION = "close_window"
EXECUTE_SCRIPT_ACTION = "execute_script"
FOCUS_ACTION = "focus"

ALL_ACTION_TYPES = [locals()[v] for v in dir() if v.endswith(ACTION_IDENTIFIER)]

# - Action Keys
ACTION_KEY = "action"
ACTION_LIST_KEY = "actions"
ALLOW_EMPTY_KEY = "allow_empty"
BYPASS_404_KEY = "bypass_404"
CHILD_KEY = "child"
CSS_SELECTOR_KEY = "css_selector"
CURRENT_URL_KEY = "current_url"
DO_NOT_INCREMENT_KEY = "do_not_increment_element_count"
DURATION_KEY = "duration"
ELEMENT_KEY = "element"
FULL_NAME_KEY = "full_name"
INDEX_KEY = "index"
INPUT_KEY = "input"
INPUT_TYPE_KEY = "input_type"
LIBRARY_KEY = "library"
PATH_KEY = "path"
POSITION_BOTTOM_KEY = "position_bottom"
POSITION_MIDDLE_KEY = "position_middle"
POSITION_TOP_KEY = "position_top"
REFERENCE_KEY = "ref"
SCRIPT_KEY = "script"
SCROLL_PADDING_KEY = "scroll_padding"
SPECIAL_KEY_KEY = "key"
SUFFIX_KEY = "suffix"
URL_KEY = "url"
VIEWPORT_ONLY_KEY = "viewport_only"
X_POSITION_KEY = "x_position"
Y_POSITION_KEY = "y_position"

ALL_ACTION_KEYS = [locals()[v] for v in dir() if v.endswith(KEY_IDENTIFIER)]


import jsonschema

from jsonschema import ValidationError, RefResolutionError


def validate(data, schema):
    """Validates json against a schema
    :param
        -   request:    JSON Dict to validate
        -   schema:     Schema to validate against
    :returns
        True if valid data
    """
    try:
        jsonschema.validate(data, schema)
    except (ValidationError, RefResolutionError) as jsonschema_exception:
        # TODO: Check for specific errors, like URL's that have a slash at the end, and give specific messages back
        message = "The given data was not valid based on the given schema: {0}".format(jsonschema_exception)
        raise SchemaValidationError(message, data, schema)
    except Exception as e:
        error_message = "Unexpected problem encountered while validating the data against the schema. " \
                        "Caught Exception: {0}".format(e)
        raise SchemaValidationError(error_message, data, schema)

    return True


class SchemaValidationError(Exception):
    def __init__(self, message=None, schema_data=None, schema=None, stacktrace=""):
        self.message = message
        self.data = schema_data if schema_data is not None else {}
        self.schema = schema if schema is not None else {}
        self.details = schema_data if schema_data is not None else {}

        self.code = 'SCHEMA_VALIDATION_ERROR'

    def __str__(self):
        return self.message

    def to_dict(self):
        self.details['data'] = self.data
        self.details['schema'] = self.schema
        return {
            'message': self.message,
            'code': self.code,
            'details': self.details
        }



import selenium_helpers
import traceback

FIELD_IDENTIFIER = "type"
STRING_FIELD = "string"
PHONE_FIELD = "phone"
ZIP_CODE_FIELD = "zip_code"
DATE_FIELD = "date"
INTEGER_FIELD = "integer"
EMAIL_FIELD = "email"
DROP_DOWN_FIELD = "drop_down"
CHECK_BOX_FIELD = "check_box"
RADIO_FIELD = "radio"
SELECT_FIELD = "select"
BUTTON_FIELD = "button"
PASSWORD_FIELD = "password"
TEXT_FIELD_TYPES = [STRING_FIELD, PHONE_FIELD, ZIP_CODE_FIELD, DATE_FIELD, INTEGER_FIELD, EMAIL_FIELD, PASSWORD_FIELD]
All_FIELD_TYPES = [DROP_DOWN_FIELD, CHECK_BOX_FIELD, RADIO_FIELD, SELECT_FIELD, BUTTON_FIELD] + TEXT_FIELD_TYPES


class FieldHandler:
    def __init__(self, selenium_helper):
        """
        This class contains methods which can be used to handle each field type.
        :param
            - selenium_helper:  SeleniumHelpers() - The current browser driver wrapper that is being interacted with.
        """
        self.sh = selenium_helper

    def dispatch_field(self, field):
        """
        Takes a 'field' dict object and dispatches it to the relevant handle method based on the 'type' key within the
        dict.
        Accepted types are:
        string  - Required keys: css_selector (string), input (string)
        phone   - Required keys: css_selector (string), input (string)
        date    - Required keys: css_selector (string), input (string)
        email   - Required keys: css_selector (string), input (string)
        integer - Required keys: css_selector (string), input (string)
        check_box - Required Keys: enum (list[dict]), input (list[int])
        radio   - Required Keys: enum (list[dict]), input (int)
        select  - Required Keys: css_selector (string), input (int), enum (list[string])
        drop_down - Required Keys: css_selector (string), input (int), enum (list[dict])
        button  - Required Keys: css_selector (string)

        :param
            - css_selector:    string - The element's css selector in the webpage's DOM.
            - input_text:      string - The text that is to be entered into the field.
            - confirm_css_selector:     bool - True if there is a repeated field used to confirm the entry into the
                                        first. This is typically most relevant to e-mail and password fields.
        """
        if FIELD_IDENTIFIER in field and field[FIELD_IDENTIFIER].lower() not in All_FIELD_TYPES:
            raise UnknownFieldType(field[FIELD_IDENTIFIER], stacktrace=traceback.format_exc())

        try:
            if field[FIELD_IDENTIFIER].lower() in TEXT_FIELD_TYPES:
                confirm_css_selector = field.get("confirm_css_selector") or None
                self.handle_text(field["css_selector"], field["input"], confirm_css_selector)

            if field[FIELD_IDENTIFIER].lower() == CHECK_BOX_FIELD:
                self.handle_check_box(field["enum"], field["input"])

            if field[FIELD_IDENTIFIER].lower() == RADIO_FIELD:
                self.handle_radio_button(field["enum"], field["input"])

            if field[FIELD_IDENTIFIER].lower() == SELECT_FIELD:
                # Default first_valid to False as it is the default.
                first_valid = field.get("first_valid") or False
                self.handle_select(field["css_selector"], field["input"], first_valid)

            if field[FIELD_IDENTIFIER].lower() == DROP_DOWN_FIELD:
                self.handle_drop_down(field["css_selector"], field["enum"], field["input"])

            if field[FIELD_IDENTIFIER].lower() == BUTTON_FIELD:
                self.handle_button(field["css_selector"])

        except FieldHandlerException as fhe:
            message = "Encountered an error dispatching the field"
            if "name" in field:
                message += " named '{0}'".format(field["name"])
            fhe.msg = "{0} | {1}".format(message, fhe.msg)
            raise fhe

        except KeyError as key:
            message = "The key {0} is missing from the field data".format(key)
            if "name" in field:
                message += " for the field named '{0}'".format(field["name"])
            message += " and so the Field Handler was unable to dispatch the field."
            raise MissingKey(message, key, stacktrace=traceback.format_exc(),
                             details={"missing_key": str(key), "field_data": field})

        except Exception as e_text:
            message = "An Unhandled Exception emerged while handling the field"
            if "name" in field:
                message += " named '{0}'".format(field["name"])
            message += " | {0}".format(e_text)
            raise FieldHandlerException(message, stacktrace=traceback.format_exc())

    def handle_text(self, css_selector="", input_text="", confirm_css_selector=None):
        """
        Logic used to fill in a text field. This method uses the selenium helpers class to interact with the field.

        :param
            - css_selector:    string - The element's css selector in the webpage's DOM.
            - input_text:      string - The text that is to be entered into the field.
            - confirm_css_selector:     bool - True if there is a repeated field used to confirm the entry into the
                                             first. This is typically most relevant to e-mail and password fields.
        """
        try:
            # - Handle the field
            self.sh.fill_an_element(css_selector=css_selector, fill_text=input_text)
            # Fill in the confirm field as well, if provided
            if confirm_css_selector:
                self.sh.fill_an_element(css_selector=confirm_css_selector, fill_text=input_text)

        except selenium_helpers.SeleniumHelperExceptions as selenium_error:
            message = "A selenium issue arose while handling the text field."
            error = SeleniumError(message, selenium_error)
            raise error

        except Exception as e_text:
            message = "An Unhandled Exception emerged while filling a Text field: {0}".format(e_text)
            raise FieldHandlerException(message)

    def handle_check_box(self, enums, input_indexes):
        """
        Logic used to fill in a check box button field. This method uses the selenium helpers class to interact with
        the field.

        :param
            - enum:          list[dict] - A list of dictionary objects. Each dictionary should contain a 'css_selector'
                                        for the element and a 'value' key representing the text that corresponds with
                                        that check box option on the form.
            - input_indexes: list[int] - A list of integers. Each integer corresponds with an index in the enum list.
                                       The dict at this index is then used to determine which css_selector is clicked.
                                       Because you can select multiple check boxes in a field you can have multiple
                                       input indexes in the list.
        """
        current_test_index = "N/A"
        try:
            # - Handle the field
            for index in input_indexes:
                current_test_index = index
                self.sh.click_an_element(css_selector=enums[index]["css_selector"])

        except KeyError as key:
            message = "Key {0} is missing from the dictionary at " \
                      "index {1} in the enum list: {2}".format(key, current_test_index, enums[current_test_index])
            raise MissingKey(message, key, stacktrace=traceback.format_exc())

        except selenium_helpers.SeleniumHelperExceptions as selenium_error:
            message = "A selenium issue arose while attempting to click the check box at the given enum's index."
            error = SeleniumError(message, selenium_error)
            raise error

        except Exception as e_text:
            message = "An Unhandled Exception emerged while filling a Check Box field: {0}".format(e_text)
            raise FieldHandlerException(message)

    def handle_radio_button(self, enums, input_index):
        """
        Logic used to fill in a radio button field. This method uses the selenium helpers class to interact with the
        field.

        :param
            - enum:         list[dict] - A list of dictionary objects. Each dictionary should contain a 'css_selector'
                                        for the element and a 'value' key representing the text that corresponds with
                                        that check box option on the form.
            - input_text:   int - This integer corresponds with an index in the enum list. The dict at this index is
                                then used to determine which css_selector is clicked.
        """
        try:
            # - Handle the field
            self.sh.click_an_element(css_selector=enums[input_index]["css_selector"])

        except KeyError as key:
            message = "Key {0} is missing from the dictionary at " \
                      "index {1} in the enum list: {2}".format(key, input_index, enums[input_index])
            raise MissingKey(message, key, stacktrace=traceback.format_exc())

        except selenium_helpers.SeleniumHelperExceptions as selenium_error:
            message = "A selenium issue arose while attempting to click the radio button at the given enum's index."
            error = SeleniumError(message, selenium_error)
            raise error

        except Exception as e_text:
            message = "An Unhandled Exception emerged while filling a Radio Button field: {0}".format(e_text)
            raise FieldHandlerException(message)

    def handle_select(self, css_selector, input_index, first_valid=False):
        """
        Logic used to fill in a select field. This method will find the select tag of the field via the css_selector
        and then choose the option that corresponds with the given index value. This method uses the selenium helpers
        class to interact with the field.

        :param
            - css_selector:     string - The <select> tag element's css selector in the webpage's DOM.
            - input_text:       int - This integer corresponds with the index of the option that will be selected from
                                    the drop down
            - first_valid       bool - True if the first option under the select tag is a valid selection.
        """
        try:
            # Create an index offset to manage the difference in Zero Base numbering between lists and :nth-child()
            index_offset = 2
            if first_valid:
                index_offset = 1

            self.sh.click_an_element(css_selector="{0} option:nth-child({1})".format(
                css_selector, input_index + index_offset))

        except selenium_helpers.SeleniumHelperExceptions as selenium_error:
            message = "A selenium issue arose while attempting to select the given select option element."
            error = SeleniumError(message, selenium_error)
            raise error

        except Exception as e_text:
            message = "An Unhandled Exception emerged while filling a Select field: {0}".format(e_text)
            raise FieldHandlerException(message)

    def handle_drop_down(self, css_selector, enums, input_index):
        """
        Logic used to fill in a drop down field. This method is used when the 'parent' element of drop down is
        separated from the options list element. The parent is clicked in order to make the option list elements
        visible. The css selector for the input element is then clicked.

        :param
            - css_selector:     string - The 'parent' element's css selector in the webpage's DOM.
            - enum:             list[dict] - A list of dictionary objects. Each dictionary should contain a
                                           'css_selector' for the element and a 'value' key representing the text that
                                           corresponds with that check box option on the form.
            - input_text:       int - This integer corresponds with the index of the option, in the enum list, that
                                    will be selected from the drop down
        """
        try:
            # - Handle the field
            # Click the parent element to reveal the options
            self.sh.click_an_element(css_selector=css_selector)
            # Click the option that corresponds with the css_selector in the given index of the enum
            self.sh.click_an_element(css_selector=enums[input_index]["css_selector"])

        except KeyError as key:
            message = "Key {0} is missing from the dictionary at " \
                      "index {1} in the enum list: {2}".format(key, input_index, enums[input_index])
            raise MissingKey(message, key)

        except selenium_helpers.SeleniumHelperExceptions as selenium_error:
            message = "A selenium issue arose while attempting to select the given select option element."
            error = SeleniumError(message, selenium_error)
            raise error

        except Exception as e_text:
            message = "An Unhandled Exception emerged while filling a Drop Down field: {0}".format(e_text)
            raise FieldHandlerException(message)

    def handle_button(self, css_selector):
        """
        Logic used to click a button field element. This can be used to navigate between form pages that have a 'Next'
        button. Or to submit the form by clicking the submit button.

        :param
            - css_selector:     string - The element's css selector in the webpage's DOM.
        """
        try:
            self.sh.click_an_element(css_selector=css_selector)

        except selenium_helpers.SeleniumHelperExceptions as selenium_error:
            message = "A selenium issue arose while attempting to select click the button"
            error = SeleniumError(message, selenium_error)
            raise error

        except Exception as e_text:
            message = "An Unhandled Exception emerged while attempting to click the button: {0}".format(e_text)
            raise FieldHandlerException(message)


class FieldHandlerException(Exception):
    def __init__(self, msg, stacktrace=None, details=None):
        self.msg = msg
        self.details = {} if details is None else details
        self.stacktrace = stacktrace
        super(FieldHandlerException, self).__init__()

    def __str__(self):
        exception_msg = "Field Handler Exception: \n"
        if self.stacktrace is not None:
            exception_msg += "{0}".format(self.stacktrace)
        if self.details:
            detail_string = "\nException Details:\n"
            for key, value in self.details.items():
                detail_string += "{0}: {1}\n".format(key, value)
            exception_msg += detail_string
        exception_msg += "Message: {0}".format(self.msg)

        return exception_msg


class MissingKey(FieldHandlerException):
    def __init__(self, message, key, stacktrace=None, details=None):
        super(MissingKey, self).__init__(msg=message, stacktrace=stacktrace, details=details)
        self.key = "{0}".format(key)
        if details:
            self.details["missing_key"] = details.get("missing_key") or key


class SeleniumError(FieldHandlerException):
    def __init__(self, message, selenium_helper_exception):
        new_message = "{0} | {1}".format(message, selenium_helper_exception.msg)
        super(SeleniumError, self).__init__(msg=new_message,
                                            stacktrace=selenium_helper_exception.stacktrace,
                                            details=selenium_helper_exception.details)


class UnknownFieldType(FieldHandlerException):
    def __init__(self, field_type, stacktrace=None):
        message = """An unknown field type of '{0}' was passed through to the field handler dispatch method.
                  Please review the field's configuration and look for typos or field types that should
                  potentially be added.""".format(field_type)
        super(UnknownFieldType, self).__init__(msg=message, stacktrace=stacktrace)
        self.field_type = "{0}".format(field_type)
        self.details["unknown_field_type"] = field_type


import unittest

from the_ark.actions import Actions, ActionException
from the_ark.field_handlers import STRING_FIELD, EMAIL_FIELD, PHONE_FIELD, ZIP_CODE_FIELD, DATE_FIELD
from the_ark.resources.action_constants import *
from the_ark.selenium_helpers import SeleniumHelpers, SeleniumHelperExceptions
from mock import patch


class ActionTestCase(unittest.TestCase):
    def setUp(self):
        self.instantiate_screenshot_class()

    @patch("the_ark.selenium_helpers.SeleniumHelpers")
    def instantiate_screenshot_class(self, selenium_helper):
        self.ac = Actions(selenium_helper)

    # - Dispatch Methods
    @patch("the_ark.actions.Actions.dispatch_action")
    def test_dispatch_list(self, mock_dispatch):
        action_list = ["pickles", "pineapples"]
        self.ac.dispatch_list_of_actions(action_list)
        mock_dispatch.return_value = True
        self.assertEqual(len(mock_dispatch.mock_calls), 2)
        mock_dispatch.assert_called_with(action_list[1], None)

    @patch("the_ark.actions.Actions.dispatch_action")
    def test_dispatch_list_without_list(self, mock_dispatch):
        action_list = False
        with self.assertRaises(ActionException) as type_error:
            self.ac.dispatch_list_of_actions(action_list)
        self.assertIn("list type", type_error.exception.msg)

    @patch("the_ark.actions.Actions.load_url")
    def test_dispatch_action(self, mock_load):
        action = {ACTION_KEY: LOAD_URL_ACTION}
        mock_load.return_value = True
        self.ac.dispatch_action(action)
        mock_load.assert_called_with(action, None)

    @patch("the_ark.actions.Actions.load_url")
    def test_dipatch_action_selenium_error(self, mock_load):
        action = {ACTION_KEY: LOAD_URL_ACTION}
        mock_load.side_effect = SeleniumHelperExceptions("Boom!", "stacktrace", "www.meltmedia.com")
        with self.assertRaises(ActionException) as selenium_error:
            self.ac.dispatch_action(action)

        self.assertIn("www.meltmedia.com", selenium_error.exception.msg)

    @patch("the_ark.actions.Actions.load_url")
    def test_dipatch_action_key_error(self, mock_load):
        action = {ACTION_KEY: LOAD_URL_ACTION}
        mock_load.side_effect = KeyError("Boom!")
        with self.assertRaises(ActionException) as key_error:
            self.ac.dispatch_action(action)

        self.assertIn("key named 'Boom!'", key_error.exception.msg)

    @patch("the_ark.actions.Actions.load_url")
    def test_dipatch_action_attribute_error(self, mock_load):
        action = {ACTION_KEY: LOAD_URL_ACTION}
        mock_load.side_effect = AttributeError("Boom!")
        with self.assertRaises(ActionException) as attr_error:
            self.ac.dispatch_action(action)

        self.assertIn("AttributeError", attr_error.exception.msg)

    @patch("the_ark.actions.Actions.load_url")
    def test_dipatch_action_general_error(self, mock_load):
        action = {ACTION_KEY: LOAD_URL_ACTION}
        mock_load.side_effect = Exception("Boom!")
        with self.assertRaises(ActionException) as e:
            self.ac.dispatch_action(action)

        self.assertIn("error", e.exception.msg)

    # - Load URL Action
    @patch("the_ark.selenium_helpers.SeleniumHelpers.load_url")
    @patch("the_ark.selenium_helpers.SeleniumHelpers.get_current_url")
    def test_load_url_action_with_just_path(self, mock_url, sh_load):
        sh = SeleniumHelpers()
        ac = Actions(sh)

        action = {
            ACTION_KEY: LOAD_URL_ACTION,
            PATH_KEY: "/pickles"
        }

        test_url = "http://www.meltmedia.com"
        mock_url.return_value = test_url
        ac.load_url(action)
        sh_load.assert_called_with(test_url + "/pickles", None)

    @patch("the_ark.selenium_helpers.SeleniumHelpers.load_url")
    def test_load_url_action_with_url_and_path(self, sh_load):
        sh = SeleniumHelpers()
        ac = Actions(sh)
        test_url = "http://www.meltmedia.com"
        action = {
            ACTION_KEY: LOAD_URL_ACTION,
            PATH_KEY: "/pickles",
            URL_KEY: test_url
        }
        ac.load_url(action)
        sh_load.assert_called_with(test_url + "/pickles", None)

    @patch("the_ark.selenium_helpers.SeleniumHelpers.load_url")
    def test_load_url_action_with_just_url_withbypass(self, sh_load):
        sh = SeleniumHelpers()
        ac = Actions(sh)
        test_url = "http://www.meltmedia.com"
        action = {
            ACTION_KEY: LOAD_URL_ACTION,
            URL_KEY: test_url,
            BYPASS_404_KEY: True
        }
        ac.load_url(action)
        sh_load.assert_called_with(test_url, True)

    # - Click Action
    def test_click(self):
        action = {
            ACTION_KEY: CLICK_ACTION,
            CSS_SELECTOR_KEY: ".pickles"
        }
        self.ac.click(action, None)

        self.ac.sh.click_an_element.assert_called_with(action[CSS_SELECTOR_KEY])

    def test_click_with_element(self):
        action = {
            ACTION_KEY: CLICK_ACTION,
            ELEMENT_KEY: True
        }
        element = "element"
        self.ac.click(action, element)

        self.ac.sh.click_an_element.assert_called_with(web_element=element)

    # - Hover Action
    def test_hover(self):
        action = {
            ACTION_KEY: HOVER_ACTION,
            CSS_SELECTOR_KEY: ".pickles"
        }
        self.ac.hover(action, None)

        self.ac.sh.hover_on_element.assert_called_with(action[CSS_SELECTOR_KEY])

    def test_hover_with_element(self):
        action = {
            ACTION_KEY: HOVER_ACTION,
            ELEMENT_KEY: True
        }
        element = "element"
        self.ac.hover(action, element)

        self.ac.sh.hover_on_element.assert_called_with(web_element=element)

    # - Enter Text Action
    def test_enter_text(self):
        action = {
            ACTION_KEY: ENTER_TEXT_ACTION,
            CSS_SELECTOR_KEY: ".enter",
            INPUT_KEY: "input text"
        }
        self.ac.enter_text(action, None)

        self.ac.sh.fill_an_element.assert_called_with(action[INPUT_KEY], action[CSS_SELECTOR_KEY])

    def test_enter_text_with_element(self):
        action = {
            ACTION_KEY: ENTER_TEXT_ACTION,
            ELEMENT_KEY: True,
            INPUT_KEY: "input text"
        }
        element = "element"
        self.ac.enter_text(action, element)

        self.ac.sh.fill_an_element.assert_called_with(action[INPUT_KEY], web_element=element)

    @patch('the_ark.input_generator.generate_string')
    def test_enter_text_as_string_with_element(self, mock_string):
        test_string = "Test String"
        mock_string.return_value = test_string
        action = {
            ACTION_KEY: ENTER_TEXT_ACTION,
            ELEMENT_KEY: True,
            INPUT_TYPE_KEY: STRING_FIELD
        }
        element = "element"
        self.ac.enter_text(action, element)

        self.ac.sh.fill_an_element.assert_called_with(test_string, web_element=element)

    @patch('the_ark.input_generator.generate_email')
    def test_enter_text_as_email(self, mock_email):
        test_email = "testingStuff@meltmedia.com"
        mock_email.return_value = test_email
        action = {
            ACTION_KEY: ENTER_TEXT_ACTION,
            CSS_SELECTOR_KEY: ".enter",
            INPUT_TYPE_KEY: EMAIL_FIELD
        }
        self.ac.enter_text(action, None)

        self.ac.sh.fill_an_element.assert_called_with(test_email, action[CSS_SELECTOR_KEY])

    @patch('the_ark.input_generator.generate_zip_code')
    def test_enter_text_as_zip_code(self, mock_zip):
        test_zip = "12345"
        mock_zip.return_value = test_zip
        action = {
            ACTION_KEY: ENTER_TEXT_ACTION,
            CSS_SELECTOR_KEY: ".enter",
            INPUT_TYPE_KEY: ZIP_CODE_FIELD
        }
        self.ac.enter_text(action, None)

        self.ac.sh.fill_an_element.assert_called_with(test_zip, action[CSS_SELECTOR_KEY])

    @patch('the_ark.input_generator.generate_phone')
    def test_enter_text_as_phone_number(self, mock_phone):
        test_phone = "5557891011"
        mock_phone.return_value = test_phone
        action = {
            ACTION_KEY: ENTER_TEXT_ACTION,
            CSS_SELECTOR_KEY: ".enter",
            INPUT_TYPE_KEY: PHONE_FIELD
        }
        self.ac.enter_text(action, None)

        self.ac.sh.fill_an_element.assert_called_with(test_phone, action[CSS_SELECTOR_KEY])

    @patch('the_ark.input_generator.generate_date')
    def test_enter_text_as_date(self, mock_date):
        test_date = "01/27/1987"
        mock_date.return_value = test_date
        action = {
            ACTION_KEY: ENTER_TEXT_ACTION,
            CSS_SELECTOR_KEY: ".enter",
            INPUT_TYPE_KEY: DATE_FIELD
        }
        self.ac.enter_text(action, None)

        self.ac.sh.fill_an_element.assert_called_with(test_date, action[CSS_SELECTOR_KEY])

    @patch('the_ark.input_generator.generate_date')
    def test_enter_text_as_unknown_type(self, mock_date):
        test_date = "01/27/1987"
        mock_date.return_value = test_date
        action = {
            ACTION_KEY: ENTER_TEXT_ACTION,
            CSS_SELECTOR_KEY: ".enter",
            INPUT_TYPE_KEY: "Pickles"
        }

        with self.assertRaises(ActionException) as unknown_input_type:
            self.ac.enter_text(action, None)

        self.assertIn(action[INPUT_TYPE_KEY], unknown_input_type.exception.msg)

    # - Scroll Actions
    def test_scroll_window_to_position(self):
        action = {
            ACTION_KEY: SCROLL_WINDOW_TO_POSITION_ACTION,
            POSITION_TOP_KEY: "1",
            POSITION_BOTTOM_KEY: "2",
            X_POSITION_KEY: 300,
            Y_POSITION_KEY: 4000
        }

        self.ac.scroll_window_to_position(action, None)

        self.ac.sh.scroll_window_to_position.assert_called_with(action[Y_POSITION_KEY], action[X_POSITION_KEY],
                                                                action[POSITION_TOP_KEY], action[POSITION_BOTTOM_KEY])

    def test_scroll_window_to_position_with_defaults(self):
        action = {
            ACTION_KEY: SCROLL_WINDOW_TO_POSITION_ACTION,
            POSITION_TOP_KEY: "1",
            X_POSITION_KEY: 300,
        }

        self.ac.scroll_window_to_position(action)

        self.ac.sh.scroll_window_to_position.assert_called_with(0, action[X_POSITION_KEY],
                                                                action[POSITION_TOP_KEY], 0)

    def test_scroll_window_to_element_defaults(self):
        action = {
            ACTION_KEY: SCROLL_WINDOW_TO_ELEMENT_ACTION,
            CSS_SELECTOR_KEY: ".scroll"
        }
        self.ac.scroll_window_to_element(action, None)

        self.ac.sh.scroll_to_element.assert_called_with(action[CSS_SELECTOR_KEY], position_bottom=None,
                                                        position_middle=None)

    def test_scroll_window_to_element_with_top_and_bottom(self):
        action = {
            ACTION_KEY: SCROLL_WINDOW_TO_ELEMENT_ACTION,
            CSS_SELECTOR_KEY: ".scroll",
            POSITION_BOTTOM_KEY: True,
            POSITION_MIDDLE_KEY: True
        }
        self.ac.scroll_window_to_element(action, None)

        self.ac.sh.scroll_to_element.assert_called_with(action[CSS_SELECTOR_KEY], position_bottom=True,
                                                        position_middle=True)

    def test_scroll_window_to_element_with_element(self):
        action = {
            ACTION_KEY: SCROLL_WINDOW_TO_ELEMENT_ACTION,
            ELEMENT_KEY: True
        }
        element = "element"
        self.ac.scroll_window_to_element(action, element)

        self.ac.sh.scroll_to_element.assert_called_with(web_element=element, position_bottom=None,
                                                        position_middle=None)

    def test_scroll_window_to_element_with_element_and_top_and_bottom(self):
        action = {
            ACTION_KEY: SCROLL_WINDOW_TO_ELEMENT_ACTION,
            ELEMENT_KEY: True,
            POSITION_BOTTOM_KEY: True,
            POSITION_MIDDLE_KEY: True
        }
        element = "element"
        self.ac.scroll_window_to_element(action, element)

        self.ac.sh.scroll_to_element.assert_called_with(position_bottom=True, position_middle=True,
                                                        web_element=element)

    def test_scroll_an_element_defaults(self):
        action = {
            ACTION_KEY: SCROLL_AN_ELEMENT_ACTION,
            CSS_SELECTOR_KEY: ".scrollable"
        }
        self.ac.scroll_an_element(action, None)

        self.ac.sh.scroll_an_element.assert_called_with(css_selector=action[CSS_SELECTOR_KEY], scroll_bottom=None,
                                                        scroll_padding=None, scroll_top=None, x_position=None,
                                                        y_position=None)

    def test_scroll_an_element_with_values(self):
        action = {
            ACTION_KEY: SCROLL_AN_ELEMENT_ACTION,
            CSS_SELECTOR_KEY: ".scrollable",
            POSITION_BOTTOM_KEY: True,
            X_POSITION_KEY: 100,
            Y_POSITION_KEY: 1200,
            SCROLL_PADDING_KEY: 45,
            POSITION_TOP_KEY: False
        }
        self.ac.scroll_an_element(action, None)

        self.ac.sh.scroll_an_element.assert_called_with(css_selector=action[CSS_SELECTOR_KEY],
                                                        scroll_bottom=action[POSITION_BOTTOM_KEY],
                                                        scroll_padding=action[SCROLL_PADDING_KEY],
                                                        scroll_top=action[POSITION_TOP_KEY],
                                                        x_position=action[X_POSITION_KEY],
                                                        y_position=action[Y_POSITION_KEY])

    def test_scroll_an_element_with_element_and_defaults(self):
        action = {
            ACTION_KEY: SCROLL_AN_ELEMENT_ACTION,
            ELEMENT_KEY: True
        }
        element = "element"
        self.ac.scroll_an_element(action, element)

        self.ac.sh.scroll_an_element.assert_called_with(web_element=element, scroll_bottom=None, scroll_padding=None,
                                                        scroll_top=None, x_position=None, y_position=None)

    def test_scroll_an_element_With_element_and_values(self):
        action = {
            ACTION_KEY: SCROLL_AN_ELEMENT_ACTION,
            ELEMENT_KEY: True,
            POSITION_BOTTOM_KEY: True,
            X_POSITION_KEY: 100,
            Y_POSITION_KEY: 1200,
            SCROLL_PADDING_KEY: 45,
            POSITION_TOP_KEY: False
        }
        element = "element"
        self.ac.scroll_an_element(action, element)

        self.ac.sh.scroll_an_element.assert_called_with(web_element=element, scroll_bottom=action[POSITION_BOTTOM_KEY],
                                                        scroll_padding=action[SCROLL_PADDING_KEY],
                                                        scroll_top=action[POSITION_TOP_KEY],
                                                        x_position=action[X_POSITION_KEY],
                                                        y_position=action[Y_POSITION_KEY])

    # - Refresh Action
    def test_refresh(self):
        self.ac.refresh("action")
        self.ac.sh.refresh.called_once()

    # - Sleep Action
    @patch('time.sleep')
    def test_sleep(self, mock_sleep):
        action = {
            ACTION_KEY: SLEEP_ACTION,
            DURATION_KEY: 10
        }
        self.ac.sleep(action)
        mock_sleep.assert_called_with(action[DURATION_KEY])

    # - Wait for Element Action
    def test_wait_for_element_defaults(self):
        action = {
            ACTION_KEY: WAIT_FOR_ELEMENT_ACTION,
            CSS_SELECTOR_KEY: ".wait"
        }
        self.ac.wait_for_element(action)
        self.ac.sh.wait_for_element.assert_called_with(action[CSS_SELECTOR_KEY], 15)

    def test_wait_for_element_with_duration(self):
        action = {
            ACTION_KEY: WAIT_FOR_ELEMENT_ACTION,
            CSS_SELECTOR_KEY: ".wait",
            DURATION_KEY: 10
        }
        self.ac.wait_for_element(action)
        self.ac.sh.wait_for_element.assert_called_with(action[CSS_SELECTOR_KEY], action[DURATION_KEY])

    # - Special Key Action
    def test_send_special_key(self):
        action = {
            ACTION_KEY: SEND_SPECIAL_KEY_ACTION,
            SPECIAL_KEY_KEY: "ENTER"
        }
        self.ac.send_special_key(action)
        self.ac.sh.send_special_key.assert_called_with(action[SPECIAL_KEY_KEY])

    # - Show Element Action
    def test_show_element(self):
        action = {
            ACTION_KEY: SHOW_ELEMENT_ACTION,
            CSS_SELECTOR_KEY: ".pickles"
        }
        self.ac.show_element(action, None)

        self.ac.sh.show_element.assert_called_with(action[CSS_SELECTOR_KEY])

    def test_show_element_with_element(self):
        action = {
            ACTION_KEY: SHOW_ELEMENT_ACTION,
            ELEMENT_KEY: True
        }
        element = "element"
        self.ac.show_element(action, element)

        self.ac.sh.show_element.assert_called_with(web_element=element)

    # - Hide Element Action
    def test_hide_element(self):
        action = {
            ACTION_KEY: HIDE_ELEMENT_ACTION,
            CSS_SELECTOR_KEY: ".pickles"
        }
        self.ac.hide_element(action, None)

        self.ac.sh.hide_element.assert_called_with(action[CSS_SELECTOR_KEY])

    def test_hide_element_with_element(self):
        action = {
            ACTION_KEY: HIDE_ELEMENT_ACTION,
            ELEMENT_KEY: True
        }
        element = "element"
        self.ac.hide_element(action, element)

        self.ac.sh.hide_element.assert_called_with(web_element=element)

    # - Execute Script Action
    def test_execute_script(self):
        action = {
            ACTION_KEY: EXECUTE_SCRIPT_ACTION,
            SCRIPT_KEY: "script text"
        }
        self.ac.execute_script(action, None)

        self.ac.sh.execute_script.assert_called_with(action[SCRIPT_KEY])

    def test_execute_script_with_element(self):
        action = {
            ACTION_KEY: EXECUTE_SCRIPT_ACTION,
            ELEMENT_KEY: True,
            SCRIPT_KEY: "script text"
        }
        element = "element"
        self.ac.execute_script(action, element)

        self.ac.sh.execute_script.assert_called_with(action[SCRIPT_KEY], element)

    # - Switch Window Handle Action
    def test_switch_window_handle(self):
        action = {
            ACTION_KEY: "switch_window_handle"
        }
        self.ac.switch_window_handle(action)
        self.ac.sh.switch_window_handle.assert_called_with()

    def test_switch_window_handle_with_index(self):
        fake_handles = ["handle1", "handle2"]
        action = {
            ACTION_KEY: "switch_window_handle",
            INDEX_KEY: 1
        }
        self.ac.sh.get_window_handles.return_value = fake_handles
        self.ac.switch_window_handle(action)
        self.ac.sh.switch_window_handle.assert_called_with(fake_handles[action[INDEX_KEY]])

    # - Close Window Action
    def test_close_window(self):
        action = {
            ACTION_KEY: CLOSE_WINDOW_ACTION
        }
        self.ac.close_window(action)
        self.ac.sh.close_window.assert_called_with()

    # - For Each Action
    @patch('the_ark.actions.Actions.dispatch_list_of_actions')
    def test_for_each(self, mock_dispatch):
        action ={
            ACTION_KEY: FOR_EACH_ACTION,
            CSS_SELECTOR_KEY: ".for-each",
            ALLOW_EMPTY_KEY: False,
            ACTION_LIST_KEY: [
                {
                    ACTION_KEY: CLICK_ACTION,
                    ELEMENT_KEY: True
                },
            ]
        }
        self.ac.sh.element_exists.return_value = True
        self.ac.sh.get_list_of_elements.return_value = ["element1", "element2"]

        self.ac.for_each(action)

        self.assertEqual(len(mock_dispatch.mock_calls), 2)

    @patch('the_ark.actions.Actions.dispatch_list_of_actions')
    def test_for_each_without_elements(self, mock_dispatch):
        action = {
            ACTION_KEY: FOR_EACH_ACTION,
            CSS_SELECTOR_KEY: ".for-each",
            ALLOW_EMPTY_KEY: False,
            ACTION_LIST_KEY: [
                {
                    ACTION_KEY: CLICK_ACTION,
                    ELEMENT_KEY: True
                },
            ]
        }
        self.ac.sh.element_exists.return_value = False

        with self.assertRaises(ActionException) as type_error:
            self.ac.for_each(action)

        self.assertIn("no elements", type_error.exception.msg)

    @patch('the_ark.actions.Actions.dispatch_list_of_actions')
    def test_for_each_child(self, mock_dispatch):
        action = {
            ACTION_KEY: FOR_EACH_ACTION,
            CSS_SELECTOR_KEY: ".for-each",
            ALLOW_EMPTY_KEY: False,
            CHILD_KEY: True,
            ACTION_LIST_KEY: [
                {
                    ACTION_KEY: CLICK_ACTION,
                    ELEMENT_KEY: True
                },
            ]
        }
        self.ac.sh.element_exists.return_value = ["element1", "element2"]
        self.ac.for_each(action)

    @patch('the_ark.actions.Actions.dispatch_list_of_actions')
    def test_for_each_allow_empty(self, mock_dispatch):
        action = {
            ACTION_KEY: FOR_EACH_ACTION,
            CSS_SELECTOR_KEY: ".for-each",
            ALLOW_EMPTY_KEY: True,
            ACTION_LIST_KEY: [
                {
                    ACTION_KEY: CLICK_ACTION,
                    ELEMENT_KEY: True
                },
            ]
        }
        self.ac.sh.element_exists.return_value = False
        self.ac.for_each(action)

    @patch('the_ark.actions.Actions.dispatch_list_of_actions')
    def test_for_each_do_not_increment(self, mock_dispatch):
        action = {
            ACTION_KEY: FOR_EACH_ACTION,
            CSS_SELECTOR_KEY: ".for-each",
            ALLOW_EMPTY_KEY: True,
            DO_NOT_INCREMENT_KEY: True,
            ACTION_LIST_KEY: [
                {
                    ACTION_KEY: CLICK_ACTION,
                    ELEMENT_KEY: True
                },
            ]
        }
        self.ac.sh.element_exists.return_value = True
        self.ac.sh.get_list_of_elements.return_value = ["element1", "element2"]
        self.ac.for_each(action)

    # - ActionException
    def test_exception_text(self):
        ac = ActionException("test")
        self.assertIn("test", str(ac))

    def test_exception_with_stacktrace(self):
        ac = ActionException("test", "stacktrace testing")
        self.assertIn("stacktrace testing", str(ac))

    def test_exception_with_details(self):
        ac = ActionException("test", details={"meltQA": "Rocks"})
        self.assertIn("Rocks", str(ac))
        self.assertIn("Exception Details", str(ac))




import unittest

from mock import patch, Mock
from the_ark import rhino_client

__author__ = 'chaley'


rhino_client_ojb = None


class UtilsTestCase(unittest.TestCase):

    def setUp(self):
        self.rhino_client_obj = rhino_client.RhinoClient('test_name',
                                                         'url', 'brand',
                                                         'branch', 'build_id',
                                                         'user',
                                                         'rhino_client_url')

    def test_set_log(self):
        self.rhino_client_obj.set_log("file_path", "link_text")
        self.assertEqual('file_path',
                         self.rhino_client_obj.test_data['result_url'])
        self.assertEqual('link_text',
                         self.rhino_client_obj.test_data['result_text'])

    @patch('requests.get')
    def test_get(self, requests_get):
        r = Mock()
        r.json.return_value = {"stuff": "stuff"}
        requests_get.return_value = r
        response = self.rhino_client_obj.get('test_id')
        self.assertEqual({"stuff": "stuff"}, response)

    @patch('requests.post')
    def test_post(self, requests_post):
        request_json = Mock()
        request_json.status_code = 201
        requests_post.return_value = request_json
        self.rhino_client_obj.post()

        self.assertEqual(True, self.rhino_client_obj.posted)

    @patch('requests.post')
    def test_post_fail(self, requests_post):
        request_json = Mock()
        request_json.status_code = 400
        requests_post.return_value = request_json

        self.assertRaises(Exception, self.rhino_client_obj.post)

    @patch('requests.put')
    def test_put(self, requests_put):
        self.rhino_client_obj.test_data['test_id'] = 156465465
        self.rhino_client_obj.posted = True

        request_json = Mock()
        request_json.status_code = 201
        request_json.json.return_value = {"stuff": "stuff"}

        requests_put.return_value = request_json
        self.rhino_client_obj.put()

        self.assertEqual(True, self.rhino_client_obj.posted)

    def test_put_posted_false(self):

        self.assertRaises(Exception, self.rhino_client_obj.put)

    @patch('requests.put')
    def test_put_status_false(self, requests_put):
        self.rhino_client_obj.test_data['test_id'] = 156465465
        self.rhino_client_obj.posted = True

        request_json = Mock()
        request_json.status_code = 500
        requests_put.return_value = request_json
        self.assertRaises(rhino_client.RhinoClientException,
                          self.rhino_client_obj.put)

    @patch('requests.post')
    def test_send_test_post(self, requests_post):
        request_json = Mock()
        request_json.status_code = 201
        requests_post.return_value = request_json
        self.rhino_client_obj.send_test("status")

        self.assertEqual(True, self.rhino_client_obj.posted)

    @patch('requests.put')
    def test_send_test_put(self, requests_put):
        self.rhino_client_obj.test_data['test_id'] = 156465465
        self.rhino_client_obj.posted = True

        request_json = Mock()
        request_json.status_code = 201

        requests_put.return_value = request_json
        self.rhino_client_obj.send_test("status")

        self.assertEqual(True, self.rhino_client_obj.posted)

if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
from datetime import datetime, timedelta
from field_handlers import DROP_DOWN_FIELD, CHECK_BOX_FIELD, RADIO_FIELD, SELECT_FIELD, BUTTON_FIELD, STRING_FIELD, \
    PHONE_FIELD, ZIP_CODE_FIELD, DATE_FIELD, INTEGER_FIELD, EMAIL_FIELD, All_FIELD_TYPES, FIELD_IDENTIFIER, PASSWORD_FIELD
import random
import string
import time
import traceback

INDEX_FIELDS = [SELECT_FIELD, DROP_DOWN_FIELD, RADIO_FIELD]
DEFAULT_STRING_MIN = 1
DEFAULT_STRING_MAX = 10
DEFAULT_INTEGER_MIN = 1
DEFAULT_INTEGER_MAX = 9
DEFAULT_INTEGER = 1
DEFAULT_INDEX_OPTIONS = 2
DEFAULT_DOMAIN = "meltmedia.com"
DEFAULT_START_DATE = str((datetime.now() - timedelta(weeks=52 * 20)).date())
DEFAULT_END_DATE = str((datetime.now() - timedelta(weeks=52 * 100)).date())
DEFAULT_DATE_FORMAT = "%m/%d/%Y"
SPECIAL_CHARACTER_LIST = ["!", "@", "#", "$", "&", "*"]


def dispatch_field(field_data, test_number=1):
    """
    Takes a 'field_data' dict object and dispatches it to the relevant input generation method based on the
    'type' key within the dict.

    :param
        - field_data:   dict - An object contain key value pairs for each of the field's parameters
        - test_number:  int - An number that specifies which submission number this generation is being used for. This
                              will help determine whether the field has been populated previously and whether to leave
                              it blank
    """
    if "type" in field_data and field_data["type"].lower() not in All_FIELD_TYPES:
        raise UnknownFieldType(field_data["type"], stacktrace=traceback.format_exc())

    try:
        required = field_data.get("required") or False
        if field_data[FIELD_IDENTIFIER].lower() == STRING_FIELD:
            return generate_string(field_data["min"], field_data["max"], test_number, required)

        elif field_data[FIELD_IDENTIFIER].lower() == INTEGER_FIELD:
            padding = field_data.get("padding") or DEFAULT_INTEGER
            return generate_integer(field_data["min"], field_data["max"], padding, test_number, required)

        elif field_data[FIELD_IDENTIFIER].lower() == EMAIL_FIELD:
            domain = field_data.get("domain") or DEFAULT_DOMAIN
            return generate_email(domain, test_number, required)

        elif field_data[FIELD_IDENTIFIER].lower() == PASSWORD_FIELD:
            return generate_password(test_number, required)

        elif field_data[FIELD_IDENTIFIER].lower() == PHONE_FIELD:
            decimal = field_data.get("decimal") or False
            parenthesis = field_data.get("parenthesis") or False
            dash = field_data.get("dash") or False
            space = field_data.get("space") or False
            return generate_phone(decimal, parenthesis, dash, space, test_number, required)

        elif field_data[FIELD_IDENTIFIER].lower() == ZIP_CODE_FIELD:
            return generate_zip_code(test_number, required)

        elif field_data[FIELD_IDENTIFIER].lower() in INDEX_FIELDS:
            always_random = field_data.get("random") or False
            return generate_index(len(field_data["enum"]), test_number, required, always_random)

        elif field_data[FIELD_IDENTIFIER].lower() == CHECK_BOX_FIELD:
            return generate_check_box(len(field_data["enum"]), test_number, required)

        else:
            # Date Field! (no elif here cause unittest coverage didn't like the branch not having an end)
            start_date = field_data.get("start_date") or DEFAULT_START_DATE
            end_date = field_data.get("end_date") or DEFAULT_END_DATE
            date_format = field_data.get("date_format") or DEFAULT_DATE_FORMAT
            return generate_date(start_date, end_date, date_format, test_number, required)

    except InputGeneratorException as ige:
        message = "Encountered an error dispatching the field"
        if "name" in field_data:
            message += " named '{0}'".format(field_data["name"])
        ige.msg = "{0} | {1}".format(message, ige.msg)
        raise ige

    except KeyError as key:
        message = "The key {0} is missing from the field data".format(key)
        if "name" in field_data:
            message += " for the field named '{0}'".format(field_data["name"])
        message += ", thus the Input Generator was unable to dispatch the field."
        raise MissingKey(message, key, stacktrace=traceback.format_exc(),
                         details={"missing_key": str(key), "field_data": field_data})

    except Exception as e_text:
        message = "An Unhandled Exception emerged while generating an input for the field"
        if "name" in field_data:
            message += " named '{0}'".format(field_data["name"])
        message += " | {0}".format(e_text)
        raise InputGeneratorException(message, stacktrace=traceback.format_exc())


def set_leave_blank(test_number, required):
    """Sets a generation method's values for whether to leave a field blank.
    :param
        -   test_number:    An int that specifies which submission number this generation is being used for. This will
                            help determine whether the field has been populated previously and whether to leave it blank
        -   required:       A bool specifying whether the field for which input if being generated is required on the
                            form on which the input will be input
    :returns
        -   bool:           Whether the input for the field should be left blank
    """
    leave_blank = False

    # If the field is not required, do logic to determine whether to leave it blank upon fill
    if required is False:
        # Always fill in the field on the first test...
        if test_number != 1:
            # ... but give 25% chance to leave blank otherwise
            leave_blank = True if random.random() < .25 else False

    return leave_blank


def check_min_vs_max(min_length, max_length):
    """
    Ensures the min is not greater than the max before creating the field value within the min - max range
    :param
        -   min_length:     The low value end of the length range
        -   max_length:     The high value end of the length range
    :return:
    """
    if min_length > max_length:
        message = "The minimum cannot be greater than the maximum value"
        raise MinGreaterThanMax(message, min_length, max_length)


def generate_string(min_length=DEFAULT_STRING_MIN, max_length=DEFAULT_STRING_MAX,
                    test_number=1, required=True):
    """ Creates a str object with a length greater than min_length and less than max_length, made up of randomly
        selected upper and lowercase letters.
    :param
        -   min_length:     The minimum length, in characters, that the generated string can be. Defaults to 1
        -   max_length:     The maximum length, in characters, that the generated string can be. Defaults to 10
        -   test_number:    An int that specifies which submission number this generation is being used for. This will
                            help determine whether the field has been populated previously and whether to leave it blank
        -   required:       A bool specifying whether the field for which input if being generated is required on the
                            form on which the input will be input
    :returns
        -   string:         The randomly generated or blank string
    """
    try:
        # Ensure the minimum and maximum values create a valid range
        check_min_vs_max(min_length, max_length)

        # Instantiate the required and leave_blank variables based on the field object and test number
        leave_blank = set_leave_blank(test_number, required)

        # Set the return to a blank string if leave_blank is true. Otherwise create a string
        if leave_blank:
            random_string = ""
        else:
            random_string = "".join(random.choice(string.ascii_letters) for num in range(random.randint(min_length,
                                                                                                        max_length)))

        return random_string

    except Exception as e_text:
        message = "Unhandled Exception caught while generating a String: {0}".format(e_text)
        raise InputGeneratorException(message)


def generate_integer(min_int=DEFAULT_INTEGER_MIN, max_int=DEFAULT_INTEGER_MAX,
                     padding=DEFAULT_INTEGER, test_number=1, required=True):
    """ Generates an str object with an int character that is greater that min_int and less than max_int.
    :param
        -   min_int:        The minimum value that the generated integer can be. Defaults to 1
        -   max_int:        The maximum value that the generated integer can be. Defaults to 9
        -   padding:        The number of characters the string will be. if the int created is 2 characters, but a
                            padding of 3 is given, then two leading zeroes are be added to make the int three characters
        -   test_number:    An int that specifies which submission number this generation is being used for. This will
                            help determine whether the field has been populated previously and whether to leave it blank
        -   required:       A bool specifying whether the field for which input if being generated is required on the
                            form on which the input will be input
    :returns
        -   integer:        The randomly generated, or blank string
    """
    try:
        # Ensure the minimum and maximum values create a valid range
        check_min_vs_max(min_int, max_int)

        # Instantiate the required and leave_blank variables based on the field object and test number
        leave_blank = set_leave_blank(test_number, required)

        # Set the return to a blank string if leave_blank is true. Otherwise create an integer
        if leave_blank:
            integer = ""
        else:
            # Create the integer value between the min and max and with the padding provided
            integer = "{0:0{1}d}".format(random.randint(min_int, max_int), padding)

        return integer

    except Exception as e_text:
        message = "Unhandled Exception caught while generating an Integer: {0}".format(e_text)
        raise InputGeneratorException(message)


def generate_email(domain=DEFAULT_DOMAIN, test_number=1, required=True):
    """ Generates a random email address in the firstname.lastname@domain format
    :param
        -   domain:         The domain address the email will be from ie. @domain. This defaults to "meltmedia.com"
        -   test_number:    An int that specifies which submission number this generation is being used for. This will
                            help determine whether the field has been populated previously and whether to leave it blank
        -   required:       A bool specifying whether the field for which input if being generated is required on the
                            form on which the input will be input
    :returns
        -   string:         The randomly generated, or blank email string
    """
    try:
        # Instantiate variables
        leave_blank = set_leave_blank(test_number, required)
        email = ""

        # Set the return to a blank string if leave_blank is true. Otherwise create an email
        if leave_blank:
            return ""
        else:
            # - Create an email in the firstname.lastname@domain format.
            #   The first and last names are generated using the generate_string method
            first_name = generate_string(6, 10)
            last_name = generate_string(6, 10)

            if domain == DEFAULT_DOMAIN:
                email = "test+"

            email += "{0}.{1}@{2}".format(first_name, last_name, domain)

            return email

    except Exception as e_text:
        message = "Unhandled Exception caught while generating an Email: {0}".format(e_text)
        raise InputGeneratorException(message)


def generate_password(test_number=1, required=True):
    """ Generates a random password with a capitol letter first and ends with a special character
    :param
        -   test_number:    An int that specifies which submission number this generation is being used for. This will
                            help determine whether the field has been populated previously and whether to leave it blank
        -   required:       A bool specifying whether the field for which input if being generated is required on the
                            form on which the input will be input
    :returns
        -   string:         The randomly generated password string
    """
    try:
        first_letter = generate_string(1, 2).upper()
        number = generate_integer(1, 2)
        body = generate_string(5, 6)
        character_position = random.randint(0, len(SPECIAL_CHARACTER_LIST) - 1)
        character = SPECIAL_CHARACTER_LIST[character_position]
        password = "{0}{1}{2}{3}".format(first_letter, body, number, character)

        return password

    except Exception as e_text:
        message = "Unhandled Exception caught while generating an Password: {0}".format(e_text)
        raise InputGeneratorException(message)


def generate_phone(decimals=False, parenthesis=False, dash=False, space=False, test_number=1, required=True):
    """ Generates a random phone number
    :param
        -   decimals:       Bool used to determine whether to put a decimals between each of the portions of the phone
                            number. When True this parameter supersedes the others as you cannot have both decimals and
                            parenthesis, or decimals and space or dash, etc.
        -   parenthesis:    Bool used to determine whether to put a parenthesis around the Area Code of the generated
                            number. The default value is False
        -   dash:           Bool used to determine whether to put a dash between the first three and last 4 digits of
                            the phone number. The default value is False
        -   space:          Bool used to determine whether to put a space between the area code and number portions
                            of the phone number. The default value is False
        -   test_number:    An int that specifies which submission number this generation is being used for. This will
                            help determine whether the field has been populated previously and whether to leave it blank
        -   required:       A bool specifying whether the field for which input if being generated is required on the
                            form on which the input will be input
    :returns
        -   string:         The randomly generated, or blank phone number
    """
    try:
        leave_blank = set_leave_blank(test_number, required)

        # Set the return to a blank string if leave_blank is true. Otherwise create a phone number
        if leave_blank:
            phone_number = ""
        else:
            # - Generate the Area Code portion of the phone number
            # Area Code should not start with 0's or 1's
            area_code = "".join(str(random.randint(2, 9)) for x in range(0, 3))

            # - Generate the number portion of the phone number
            # The first three digits of the number should not start with 0's or 1's
            start = "".join(str(random.randint(2, 9)) for x in range(0, 3))
            finish = "".join(str(random.randint(0, 9)) for y in range(0, 4))

            # - Format the number
            # Use only the decimal formatting if the user specified they wanted decimals
            if decimals:
                phone_number = "{0}.{1}.{2}".format(area_code, start, finish)
            else:
                # Surround the area code in parenthesis if parenthesis parameter is True
                area_code = "({0})".format(area_code) if parenthesis else area_code

                # Format the "number" portion of the phone number
                # Dash takes precedence over space
                if dash:
                    if not space and not parenthesis:
                        number = "-{0}-{1}".format(start, finish)
                    else:
                        # Add the dash between the start and finish of the number if dash parameter is True
                        number = "{0}-{1}".format(start, finish)
                elif space:
                    # Add a space if space parameter is True, but dash is False
                    number = "{0} {1}".format(start, finish)
                else:
                    number = "{0}{1}".format(start, finish)

                # Stitch the area code and number together
                if space:
                    phone_number = "{0} {1}".format(area_code, number)
                else:
                    phone_number = area_code + number

        return phone_number

    except Exception as e_text:
        message = "Unhandled Exception caught while generating a Phone Number: {0}".format(e_text)
        raise InputGeneratorException(message)


def generate_zip_code(test_number=1, required=True):
    """ Generates a random 5 digit string to act as a ZIP code
    :param
        -   test_number:    An int that specifies which submission number this generation is being used for. This will
                            help determine whether the field has been populated previously and whether to leave it blank
        -   required:       A bool specifying whether the field for which input if being generated is required on the
                            form on which the input will be input
    :returns
        -   sting:          The randomly generated, or blank ZIP Code string
    """
    try:
        leave_blank = set_leave_blank(test_number, required)

        # Set the return to a blank string if leave_blank is true. Otherwise create a zip code
        if leave_blank:
            zip_code = ""
        else:
            # - Generate the ZIP Code
            # Prevent the first number from being 0 so that it is not truncated when the output is viewed in Excel
            first = str(random.randint(1, 9))
            # The last four digits are between 0 and 9
            zip_code = first + "".join(str(random.randint(0, 9)) for z in range(0, 4))

        return zip_code

    except Exception as e_text:
        message = "Unhandled Exception caught while generating an Zip Code: {0}".format(e_text)
        raise InputGeneratorException(message)


def generate_index(num_of_options=DEFAULT_INDEX_OPTIONS, test_number=1, required=True, always_random=False):
    """ Calculates which option should be selected from the given field based on test number. The index is randomly
        selected after all options have been used at least once.
    :param
        -   num_of_options: The number oof options available in the list. If a field object is passed in then this
                            value is overwritten by the "enum" key for that field. Defaults to a value of 2
        -   test_number:    An int that specifies which submission number this generation is being used for. This will
                            help determine whether all options for the field have been used or not and whether to leave
                            it blank
        -   required:       A bool specifying whether the field for which input if being generated is required on the
                            form on which the input will be input
        -   always_random:  A bool specifying whether to always select a random index for this field. This will
                            typically be used for fields like State which have 50+ options. In these cases we want to
                            select a new item each time, but are not necessarily all options or in any particular order
    :returns
        -   integer:        The randomly generated (int), or blank index (string) to select from the field's options
    """
    try:
        leave_blank = set_leave_blank(test_number, required)

        # Set test_number to a default of 1 unless a value was passed in.
        test_number = 1 if not test_number else test_number

        random_choice = None
        all_options_used = True if test_number > num_of_options else False

        # - Always choose a random index if all options have already been sent once
        if all_options_used:
            random_choice = True

        # - Generate the input index to use when filling out this field
        if leave_blank:
            # Set the return to a blank string if leave_blank is true.
            input_index = ""
        elif random_choice or always_random:
            # Select a random index if the field is "random" or if all options have already been selected previously
            input_index = random.randint(0, num_of_options - 1)
        else:
            # Select the modulo unless the test_number and number of options is the same, then select the last index
            if test_number == num_of_options:
                input_index = num_of_options - 1
            else:
                input_index = test_number % num_of_options - 1

        return input_index

    except Exception as e_text:
        message = "Unhandled Exception caught while generating an input index: {0}".format(e_text)
        raise InputGeneratorException(message)


def generate_check_box(num_of_options=1, test_number=1, required=True):
    """ Calculates which check box from the list of boxes should be selected based on test number. The index is
        randomly selected after all options have been used at least once. When randomly selecting there is a chance
        that more than one of the checkboxes will be selected.
    :param
        -   num_of_options: The number oof options available in the list. If a field object is passed in then this
                            value is overwritten by the "enum" key for that field. Defaults to a value of 1
        -   test_number:    An int that specifies which submission number this generation is being used for. This will
                            help determine whether all options for the field have been used or not and whether to leave
                            it blank
        -   required:       A bool specifying whether the field for which input if being generated is required on the
                            form on which the input will be input
    :returns
        -   list:           A list of indexes (integers) to select from the check box list when filling out this form
    """
    try:
        leave_blank = False
        random_choice = None
        all_options_used = True if test_number > num_of_options else False

        # 25% chance of leaving the field blank if it is not required and all options have already been used
        if required is False and all_options_used:
            if random.random() < .25:
                leave_blank = True

        # Always select a random index if all options have already been used
        if all_options_used:
            random_choice = True

        input_indexes = []

        if leave_blank:
            pass
        elif random_choice and num_of_options > 1:
            # - Add at least one item to the input list
            index = random.randint(0, num_of_options - 1)
            input_indexes.append(index)

            # - 50/50 chance of adding another input index to the list
            while random.random() <= .5 and len(input_indexes) < num_of_options:
                # Cycle through indexes until finding one that is not already being used
                while index in input_indexes:
                    index = random.randint(0, num_of_options - 1)
                # Add the index to the list
                input_indexes.append(index)
        else:
            # Select the modulo unless the test_number and number of options is the same, then select the last index
            if test_number == num_of_options:
                input_indexes.append(num_of_options - 1)
            else:
                input_indexes.append(test_number % num_of_options - 1)

        return input_indexes

    except Exception as e_text:
        message = "Unhandled Exception caught while generating the Check Box input indexes: {0}".format(e_text)
        raise InputGeneratorException(message)


def generate_date(start_date=DEFAULT_START_DATE, end_date=DEFAULT_END_DATE, date_format=DEFAULT_DATE_FORMAT,
                  test_number=1, required=True):
    """ Generates a date in the given date format. By default this date will be -18 to -100 years ago in order to keep
        the user over the age of 18 when filling out birthdays. However these dates can be passed through if need be.
    :param
        -   start_date:     Date in "%Y-%m-%d" format or an integer specifying days. This is the default format for
                            datetime.now().date(). This date is used as the furthest back in history the generated date
                            will be. If the user sends through an integer rather than a date the integer will be used
                            to specify how many days in the past to start generating from. If the integer value is
                            negative, then the start date will be in the future. If no start date is given the date
                            100 years, or 52*100 weeks, ago will be used.
        -   end_date:       Date in "%Y-%m-%d" format or an integer specifying days. This is the default format for
                            datetime.now().date(). This date is used as the most recent day in history the generated
                            date will be. If the user sends through an integer rather than a date the integer will be
                            used to specify how many days in the past to stop generating from. If the integer value is
                            negative, then the end date will be in the future. If no end date is given an end date of
                            18 years, or 52*18 weeks, ago will be used.
        -   date_format:    The format you'd like the date to be when it is returned. This defaults to %m/%d/%Y which
                            would make a MM/DD/YYYY format. Other format examples are %y-%m-%d to get YY-MM-DD or
                            %d%m%Y to get DDMMYYYY, etc.
        -   test_number:    An int that specifies which submission number this generation is being used for. This will
                            help determine whether all options for the field have been used or not and whether to leave
                            it blank
        -   required:       A bool specifying whether the field for which input if being generated is required on the
                            form on which the input will be input
    :returns
        -   string:         A string formatted to look like a date
    """
    try:
        leave_blank = set_leave_blank(test_number, required)

        if leave_blank:
            return ""
        else:
            if isinstance(start_date, int):
                start_date = str((datetime.now() - timedelta(days=start_date)).date())
            if isinstance(end_date, int):
                end_date = str((datetime.now() - timedelta(days=end_date)).date())

            start_time = time.mktime(time.strptime(start_date, "%Y-%m-%d"))
            end_time = time.mktime(time.strptime(end_date, "%Y-%m-%d"))

            random_time = start_time + random.random() * (end_time - start_time)

            return time.strftime(date_format, time.localtime(random_time))

    except Exception as e_text:
        message = "Unhandled Exception caught while generating a Date: {0}".format(e_text)
        raise InputGeneratorException(message)


class InputGeneratorException(Exception):
    def __init__(self, msg, stacktrace=None, details=None):
        self.msg = msg
        self.details = {} if details is None else details
        self.stacktrace = stacktrace
        super(InputGeneratorException, self).__init__()

    def __str__(self):
        exception_msg = "Input Generator Exception: \n"
        if self.stacktrace is not None:
            exception_msg += "{0}".format(self.stacktrace)
        if self.details:
            detail_string = "\nException Details:\n"
            for key, value in self.details.items():
                detail_string += "{0}: {1}\n".format(key, value)
            exception_msg += detail_string
        exception_msg += "Message: {0}".format(self.msg)

        return exception_msg


class MissingKey(InputGeneratorException):
    def __init__(self, message, key, stacktrace=None, details=None):
        super(MissingKey, self).__init__(msg=message, stacktrace=stacktrace, details=details)
        self.key = "{0}".format(key)
        self.details["missing_key"] = key if details is None or "missing_key" not in details else details["missing_key"]


class MinGreaterThanMax(InputGeneratorException):
    def __init__(self, message, minimum, maximum, stacktrace=None, details=None):
        super(MinGreaterThanMax, self).__init__(msg=message, stacktrace=stacktrace, details=details)
        self.details["min"] = minimum if details is None or "min" not in details else details["min"]
        self.details["max"] = maximum if details is None or "max" not in details else details["max"]


class UnknownFieldType(InputGeneratorException):
    def __init__(self, field_type, stacktrace=None):
        message = """An unknown field type of '{0}' was passed through to the input generator dispatch method.
                  Please review the field's configuration and look for typos or field types that should
                  potentially be added.""".format(field_type)
        super(UnknownFieldType, self).__init__(msg=message, stacktrace=stacktrace)
        self.field_type = "{0}".format(field_type)
        self.details["unknown_field_type"] = field_type


import unittest
from mock import patch, Mock

from the_ark import picard_client

picard_client_obj = None


class UtilsTestCase(unittest.TestCase):

    def setUp(self):
        self.picard_client_obj = picard_client.PicardClient()

    @patch('requests.post')
    def test_send_to_picard(self, requests_post):
        r = Mock()
        r.status_code = 200
        r.text = '{"first_name": "bill"}'
        requests_post.return_value = r
        self.assertEqual(self.picard_client_obj.send_to_picard
                         ('url', 'data',
                          self.picard_client_obj.create_headers()),
                         {"first_name": "bill"})
        self.assertEqual(self.picard_client_obj.send_to_picard
                         ('url', 'data'),
                         {"first_name": "bill"})

    @patch('requests.post')
    def test_send_to_picard_400(self, requests_post):
        r = Mock()
        r.status_code = 400
        r.text = 'tish broke'
        requests_post.return_value = r

        self.assertRaises(picard_client.PicardClientException, self.picard_client_obj.send_to_picard, '', '', '')

        r.status_code = 303
        self.assertRaises(picard_client.PicardClientException, self.picard_client_obj.send_to_picard, '', '', '')



import logging
import json
import requests


class PicardClient(object):

    def __init__(self):
        """
        Contains methods used to query Epsilon.
        """

    def send_to_picard(self, schema_url, form_data, headers=None, **kwargs):
        r = None

        if not headers:
            headers = self.create_headers()

        r = requests.post(schema_url, form_data, headers=headers, **kwargs)

        # Able to reach Picard, but request failed
        if r.status_code == 400:
            raise PicardClientException(r.text)
        # Unable to reach picard
        if r.status_code is not 200:
            raise PicardClientException("Unable to Post to Picard")

        # Received a valid response from Picard. Return the data received.
        return json.loads(r.text)

    def create_headers(self, referer="qa.picardclient.com"):
        return {"Referer": referer}


class PicardClientException(Exception):
    def __init__(self, arg):
        self.msg = arg


import logging

import requests
import urlparse
from selenium import common
from selenium import webdriver
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as expected_condition
from selenium.webdriver.firefox.firefox_binary import FirefoxBinary
import traceback


class SeleniumHelpers:

    def __init__(self):
        """
        Methods to do various and repeatable selenium tasks.
        """
        self.log = logging.getLogger(self.__class__.__name__)
        self.driver = None
        self.desired_capabilities = {}

    def create_driver(self, **desired_capabilities):
        """
        Creating a driver with the desired settings.
        :param
            -   desired_capabilities:   dictionary - Settings used to set up the desired browser.
        """
        try:
            if desired_capabilities.get("username") and desired_capabilities.get("access_key"):
                sauce_url = "http://{0}:{1}@ondemand.saucelabs.com:80/wd/hub".format(desired_capabilities["username"],
                                                                                     desired_capabilities["access_key"])
                self.driver = webdriver.Remote(desired_capabilities=desired_capabilities, command_executor=sauce_url)
            elif desired_capabilities.get("mobile"):
                self.driver = webdriver.Remote(desired_capabilities=desired_capabilities)
            elif desired_capabilities.get("browserName").lower() == "chrome":
                self.driver = webdriver.Chrome()
            elif desired_capabilities.get("browserName").lower() == "firefox":
                binary = FirefoxBinary(desired_capabilities["binary"]) if "binary" in desired_capabilities else None
                self.driver = webdriver.Firefox(firefox_binary=binary)
            elif desired_capabilities.get("browserName").lower() == "phantomjs":
                binary_path = desired_capabilities.get("binary", "phantomjs")
                self.driver = webdriver.PhantomJS(binary_path)
            elif desired_capabilities.get("browserName").lower() == "safari":
                self.driver = webdriver.Safari()
            else:
                message = "No driver has been created. Pass through the needed desired capabilities in order to " \
                          "create a driver. | Desired Capabilities: {0}".format(desired_capabilities)
                raise DriverAttributeError(msg=message)

            # Set the desired_capabilities variable on the class if the browser creation was successful
            self.desired_capabilities = desired_capabilities

            return self.driver
        except Exception as driver_creation_error:
            message = "There was an issue creating a driver with the specified desired capabilities: {0}\n" \
                      "<{1}>".format(desired_capabilities, driver_creation_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def resize_browser(self, width=None, height=None):
        """
        This will resize the browser with the given width and/or height.
        :param
            -   width:  integer - The number the width of the browser will be re-sized to.
            -   height: integer - The number the height of the browser will be re-sized to.
        """
        try:
            if not width and not height:
                pass
            elif width and height:
                self.driver.set_window_size(width, height)
            elif width:
                self.driver.set_window_size(width, self.driver.get_window_size()["height"])
            else:
                self.driver.set_window_size(self.driver.get_window_size()["width"], height)
        except Exception as resize_error:
            message = "Unable to resize the browser with the given width ({0}) and/or height ({1}) value(s)\n" \
                      "<{2}>".format(width, height, resize_error)
            raise DriverSizeError(msg=message, stacktrace=traceback.format_exc(), width=width, height=height)

    def get_window_size(self, get_only_width=False, get_only_height=False):
        """
        This will return the width and/or height of the full window. These numbers will include the address bar,
        favorites bar, etc.
        :param
            -   get_only_width: boolean - Whether or not to just return the width of the window.
            -   get_only_height:    boolean - Whether or not to just return the height of the window.
        :return
            -   window_width: integer - The number the width of the window is at.
            -   window_height:    integer - The number the height of the window is at.
        """
        try:
            window_size = self.driver.get_window_size()
            if get_only_width and not get_only_height:
                window_width = window_size["width"]
                return window_width
            elif get_only_height and not get_only_width:
                window_height = window_size["height"]
                return window_height
            else:
                return window_size["width"], window_size["height"]
        except Exception as get_window_size_error:
            message = "Unable to get the width and/or the height of the window.\n" \
                      "<{0}>".format(get_window_size_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def load_url(self, url, bypass_status_code_check=False):
        """
        This will check to see if the status code of the URL is not 4XX or 5XX and navigate to the URL. If the
        bypass_status_code_check is set to True it will just navigate to the given URL.
        :param
            -   url:    string - A valid URL (e.g. "http://www.google.com")
            -   bypass_status_code_check:   boolean - Navigate to the given URL without checking the status code or not.
        """
        try:
            if bypass_status_code_check:
                self.driver.get(url)
            else:
                url_request = requests.get(url)
                if url_request.status_code == requests.codes.ok:
                    self.driver.get(url)
                else:
                    message = "The URL: {0} has the status code of: {1}. You may bypass the status code check if you " \
                              "need to navigate to this URL.".format(url, url_request.status_code)
                    raise DriverURLError(msg=message, desired_url=url)
        except Exception as get_url_error:
            message = "Unable to navigate to the desired URL: {0}\n" \
                      "<{1}>".format(url, get_url_error)
            raise DriverURLError(msg=message, stacktrace=traceback.format_exc(), desired_url=url)

    def get_current_url(self):
        """
        This will get and return the URL the driver is currently on.
        :return
            -   current_url:  string - The current URL the driver is on.
        """
        try:
            current_url = self.driver.current_url
            return current_url
        except Exception as get_current_url_error:
            message = "Unable to get the URL the driver is currently on.\n" \
                      "<{0}>".format(get_current_url_error)
            raise DriverURLError(msg=message, stacktrace=traceback.format_exc())

    def refresh_driver(self):
        """
        This will refresh the page the driver is currently on.
        """
        try:
            self.driver.refresh()
        except Exception as refresh_driver_error:
            message = "Unable to refresh the driver.\n" \
                      "<{0}>".format(refresh_driver_error)
            raise DriverURLError(msg=message, stacktrace=traceback.format_exc())

    def get_viewport_size(self, get_only_width=False, get_only_height=False):
        """
        This will get the width and/or height of the viewport. The reason for not using driver.get_window_size here
        instead is because it's not just getting the height of the viewport but the whole window (address bar,
        favorites bar, etc.). In order to be accurate this uses the execute_script method with scripts to get the
        clientWidth and/or clientHeight.
        :param
            -   get_only_width: boolean - Whether or not to just return the width of the viewport.
            -   get_only_height:    boolean - Whether or not to just return the height of the viewport.
        :return
            -   viewport_width: integer - The number the width of the viewport is at.
            -   viewport_height:    integer - The number the height of the viewport is at.
        """
        try:
            viewport_width = self.execute_script("return document.documentElement.clientWidth;")
            viewport_height = self.execute_script("return document.documentElement.clientHeight;")
            if get_only_width and not get_only_height:
                return viewport_width
            elif get_only_height and not get_only_width:
                return viewport_height
            else:
                return viewport_width, viewport_height
        except Exception as get_viewport_size_error:
            message = "Unable to get the width and/or the height of the viewport.\n" \
                      "<{0}>".format(get_viewport_size_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def get_window_handles(self, get_current=None):
        """
        This will get and return a list of windows or tabs currently open.
        :return
            -   current_handle:  unicode - The current window handle of the driver.
            -   window_handles:    list - A list of the current windows or tabs open in the driver.
        """
        try:
            if get_current:
                current_handle = self.driver.current_window_handle
                return current_handle
            else:
                window_handles = self.driver.window_handles
                return window_handles
        except Exception as get_handle_error:
            message = "Unable to get window handle(s).\n<{0}>".format(get_handle_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def switch_window_handle(self, specific_handle=None):
        """
        This will either switch to a specified window handle or the latest window handle.
        :param
            -   specific_handle:    unicode - The specific window handle to switch to in the driver.
        """
        try:
            if specific_handle:
                self.driver.switch_to.window(specific_handle)
            else:
                window_handles = self.get_window_handles()
                self.driver.switch_to.window(window_handles[-1])
        except Exception as window_handle_error:
            message = "Unable to switch window handles.\n<{0}>".format(window_handle_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def get_screenshot_base64(self):
        """
        Get image data as base64 of the current page.
        :return
            -   base64_image:  base64 - Image data of the current page.
        """
        try:
            base64_image = self.driver.get_screenshot_as_base64()
            return base64_image
        except Exception as base64_error:
            message = "Unable to get screenshot as base64. The browser might have been closed.\n" \
                      "<{0}>".format(base64_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def save_screenshot_as_file(self, file_path, file_name):
        """
        Capture a screenshot of the current page. The file path and file name are both required. The file_name needs to
        include the file extension.
        :param
            -   file_path:  string - The path the image will be saved to.
            -   file_name:  string = The name that the image will be saved with, including the extension.
        """
        try:
            self.driver.get_screenshot_as_file(file_path + file_name)
        except TypeError as screenshot_error:
            message = "Unable to save screenshot '{0}' to '{1}' on: {2}\n<{3}>".format(file_name, file_path,
                                                                                       self.driver.current_url,
                                                                                       screenshot_error)
            raise ScreenshotError(msg=message, stacktrace=traceback.format_exc(), current_url=self.driver.current_url,
                                  file_name=file_name, file_path=file_path)
        except Exception as unexpected_error:
            message = "Unable to save screenshot '{0}' to '{1}'. The browser might have been closed.\n" \
                      "<{2}>".format(file_name, file_path, unexpected_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def close_window(self):
        """
        This will close the active window of the driver.
        """
        try:
            self.driver.close()
        except Exception as close_error:
            message = "Unable to close the current window. Is it possible you already closed the window?\n" \
                      "<{0}>".format(close_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def quit_driver(self):
        """
        This will quit the driver.
        """
        try:
            self.driver.quit()
        except Exception as quit_error:
            message = "Unable to quit the driver. Is it possible you already quit the driver?\n" \
                      "<{0}>".format(quit_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def element_exists(self, css_selector):
        """
        This will ensure that an element exists on the page under test, if not an exception will be raised.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
        """
        try:
            self.driver.find_element_by_css_selector(css_selector)
            return True
        except common.exceptions.NoSuchElementException as no_such:
            return False

    def ensure_element_visible(self, css_selector=None, web_element=None):
        """
        This will ensure that an element is visible, if not an exception will be raised.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
        """
        if web_element:
            element_visible = web_element.is_displayed()
        else:
            if self.element_exists(css_selector):
                element_visible = self.driver.find_element_by_css_selector(css_selector).is_displayed()
            else:
                message = "Element '{0}' does not exist on page '{1}'.".format(css_selector, self.driver.current_url)
                raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                                   current_url=self.driver.current_url, css_selector=css_selector)
        if not element_visible:
            message = "The element is not visible on page '{0}'. | CSS Selector: '{1}' or WebElement passed through."\
                .format(self.driver.current_url, css_selector)
            raise ElementNotVisibleError(msg=message, stacktrace=traceback.format_exc(),
                                         current_url=self.driver.current_url, css_selector=css_selector)
        else:
            return element_visible

    def get_element(self, css_selector):
        """
        Find a specific element on the page using a css_selector.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
        :return
            -   web_element:    object - The WebElement object that has been found.
        """
        try:
            web_element = self.driver.find_element_by_css_selector(css_selector)
            return web_element
        except common.exceptions.NoSuchElementException as no_such:
            message = "Element '{0}' does not exist on page '{1}' and could not be returned.\n" \
                      "<{2}>".format(css_selector, self.driver.current_url, no_such)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)
        except Exception as unexpected_error:
            message = "Unable to find and return the element '{0}' on page '{1}'.\n" \
                      "<{2}>".format(css_selector, self.driver.current_url, unexpected_error)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def get_list_of_elements(self, css_selector):
        """
        Return a full list of elements from a drop down menu, checkboxes, radio buttons, etc.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
        :return
            -   list_of_elements:   list - The full list of web elements from a parent selector (e.g. drop down menus)
        """
        if self.element_exists(css_selector):
            list_of_elements = self.driver.find_elements_by_css_selector(css_selector)
            return list_of_elements
        else:
            message = "Element '{0}' does not exist on page '{1}'.".format(css_selector, self.driver.current_url)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def wait_for_element(self, css_selector, wait_time=15, visible=False):
        """
        This will wait for a specific element to be present on the page within a specified amount of time, in seconds.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   wait_time:      integer - The amount of time, in seconds, given to wait for an element to be present.
            -   visible:        boolean - If true, wait for the element to be visible on the page; present otherwise
        """
        try:
            if visible:
                # Wait for element to be visible on the page
                WebDriverWait(
                    self.driver, wait_time).until(expected_condition.visibility_of_element_located((By.CSS_SELECTOR,
                                                                                                    css_selector)))
            else:
                # Wait for element to be present on the page
                WebDriverWait(
                    self.driver, wait_time).until(expected_condition.presence_of_element_located((By.CSS_SELECTOR,
                                                                                                  css_selector)))
        except common.exceptions.TimeoutException as timeout:
            message = "Element '{0}' does not exist on page '{1}' after waiting {2} seconds.\n" \
                      "<{3}>".format(css_selector, self.driver.current_url, wait_time, timeout)
            raise TimeoutError(msg=message, stacktrace=traceback.format_exc(), current_url=self.driver.current_url,
                               css_selector=css_selector, wait_time=wait_time)

    def click_an_element(self, css_selector=None, web_element=None):
        """
        This will click an element on a page.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.ensure_element_visible(web_element=web_element, css_selector=css_selector)
            web_element.click()
        except SeleniumHelperExceptions as click_error:
            click_error.msg = "Unable to click element. | Based off the CSS Selector: '{0}' or WebElement " \
                              "passed through.".format(css_selector)
            raise click_error
        except Exception as unexpected_error:
            message = "An unexpected error occurred attempting to click the element on page '{0}'.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)

            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def click_element_with_offset(self, css_selector=None, web_element=None, x_position=0, y_position=0):
        """
        Click an element with an offset. The offset is relative to the top-left corner of the specified element.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
            -   y_position: integer - The position at which the mouse will be placed vertically.
            -   x_position: integer - The position at which the mouse will be placed horizontally.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.ensure_element_visible(web_element=web_element, css_selector=css_selector)
            ActionChains(self.driver).move_to_element_with_offset(web_element, x_position, y_position).click().perform()
        except SeleniumHelperExceptions as click_location_error:
            click_location_error.msg = "Unable to click the position ({0}, {1}). | " \
                                       "Based off the CSS Selector: '{2}' or WebElement passed through. | "\
                                           .format(x_position, y_position, css_selector) + click_location_error.msg
            raise click_location_error
        except Exception as unexpected_error:
            message = "Unable to click at the position ({0}, {1}) of the element on page '{2}'.\n" \
                      "<{3}>".format(x_position, y_position, self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ClickPositionError(msg=message, stacktrace=traceback.format_exc(),
                                     current_url=self.driver.current_url, css_selector=css_selector,
                                     y_position=y_position, x_position=x_position)

    def double_click(self, css_selector=None, web_element=None):
        """
        Double click an element on the page.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.ensure_element_visible(web_element=web_element, css_selector=css_selector)
            ActionChains(self.driver).double_click(web_element).perform()
        except SeleniumHelperExceptions as double_click_error:
            double_click_error.msg = "Unable to double click element. | " + double_click_error.msg
            raise double_click_error
        except Exception as unexpected_error:
            message = "Unable to double-click the element on page '{0}'.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def move_cursor_to_location(self, x_position=0, y_position=0, click=False):
        """
        Move the cursor to a specific location on the page. This will first move the cursor to (0,0) on the page and
        then move to the coordinates the user passes in. There is an option to click after the cursors moves to the
        specified coordinates.
        :param
            -   y_position: integer - The position at which the mouse will be placed vertically.
            -   x_position: integer - The position at which the mouse will be placed horizontally.
            -   click: boolean - Whether or not a click will be performed after the the cursor is moved.
        """
        try:
            # Starting the cursor off at the absolute (0,0) point on the page.
            web_element = self.get_element("body")
            ActionChains(self.driver).move_to_element_with_offset(web_element, 0, 0).perform()
            if click:
                # Moving the cursor to and clicking the specified coordinates provided.
                ActionChains(self.driver).move_by_offset(x_position, y_position).click().perform()
            else:
                # Moving the cursor to the specified coordinates provided.
                ActionChains(self.driver).move_by_offset(x_position, y_position).perform()
        except Exception as unexpected_error:
            message = "Unable to move the cursor to ({0},{1}) on page '{2}'.\n" \
                      "<{3}>".format(x_position, y_position, self.driver.current_url, unexpected_error)
            raise CursorLocationError(msg=message, stacktrace=traceback.format_exc(),
                                      current_url=self.driver.current_url, x_position=x_position, y_position=y_position)

    def clear_an_element(self, css_selector=None, web_element=None):
        """
        This will clear a field on a page.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.click_an_element(web_element=web_element, css_selector=css_selector)
            web_element.clear()
        except SeleniumHelperExceptions as clear_error:
            clear_error.msg = "Unable to clear element. | " + clear_error.msg
            raise clear_error
        except Exception as unexpected_error:
            message = "Unable to clear the element on page '{0}'.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def fill_an_element(self, fill_text, css_selector=None, web_element=None):
        """
        This will fill a field on a page.
        :param
            -   fill_text:  string - The text that will be sent through to a field on page.
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.clear_an_element(web_element=web_element, css_selector=css_selector)
            web_element.send_keys(fill_text)
        except SeleniumHelperExceptions as fill_error:
            fill_error.msg = "Unable to fill element. | " + fill_error.msg
            raise fill_error
        except Exception as unexpected_error:
            message = "Unable to fill the element on page '{0}'.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def send_special_key(self, special_key):
        """
        This will send a special key through to the driver (e.g. TAB, ENTER, RETURN).
        :param
            -   special_key:  string - The key that will be sent through to the driver.
        """
        try:
            ActionChains(self.driver).send_keys(getattr(Keys, special_key.upper())).perform()
        except Exception as send_special_key_error:
            message = "Unable to send the special key '{0}'.\n" \
                      "<{1}>".format(special_key, send_special_key_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def hover_on_element(self, css_selector=None, web_element=None):
        """
        This will hover over an element on a page.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.ensure_element_visible(web_element=web_element, css_selector=css_selector)
            hover = ActionChains(self.driver).move_to_element(web_element)
            hover.perform()
        except SeleniumHelperExceptions as hover_error:
            hover_error.msg = "Unable to hover over element. | " + hover_error.msg
            raise hover_error
        except Exception as unexpected_error:
            message = "Unable to hover over the element on page '{0}'.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def execute_script(self, script, *script_arguments):
        """
        This will run JavaScript through the browser.
        :param
            -   script: string - JavaScript to be sent through to browser.
            -   argument: list - Arguments to be used in the script that is sent to the browser.
        :return
            -   script: string, int - Any value that the script could return for future consumption.
        """
        try:
            if script_arguments:
                # Execute the script with a specific WebElement.
                return self.driver.execute_script(script, *script_arguments)
            else:
                # Execute the script without a specific WebElement.
                return self.driver.execute_script(script)
        except Exception as unexpected_error:
            message = "Unable to execute given script on page '{0}'.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def scroll_to_element(self, css_selector=None, web_element=None, position_bottom=False, position_middle=False):
        """
        This will scroll to an element on a page. This element can be put at the top, the bottom, or the middle
        (or close to) of the page.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
            -   position_top:   boolean - Whether or not the element will be at the top of the page.
            -   position_bottom:    boolean - Whether or not the element will be at the bottom of the page.
            -   position_middle:    boolean - Whether or not the element will be in the middle of the page.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.ensure_element_visible(web_element=web_element, css_selector=css_selector)
            if position_bottom:
                # Scroll the window so the bottom of the element will be at the bottom of the window.
                self.execute_script("var element = arguments[0]; element.scrollIntoView(false);",
                                    web_element)
            elif position_middle:
                # Find the scroll position of the top of the element
                element_position = self.execute_script("var element = arguments[0]; "
                                                       "var height = element.offsetTop; "
                                                       "return height", web_element)
                # Determine the position that is half a screen height above the element
                screen_padding = (self.driver.get_window_size()["height"] / 2)
                scroll_position = element_position - screen_padding
                # Scroll to that position
                self.execute_script("window.scrollTo(0, arguments[0]);", scroll_position)
            else:
                # Scroll the window so the top of the element will be at the top of the window.
                self.execute_script("var element = arguments[0]; element.scrollIntoView(true);",
                                    web_element)

        except SeleniumHelperExceptions as scroll_to_element_error:
            scroll_to_element_error.msg = "Unable to scroll to element. | " + scroll_to_element_error.msg
            raise scroll_to_element_error
        except Exception as unexpected_error:
            message = "Unable to scroll to the element on page '{0}'.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def scroll_window_to_position(self, y_position=0, x_position=0, scroll_top=False, scroll_bottom=False):
        """
        This will scroll to a specific position on the current page, scroll to the top of the page, or it will scroll
        to the bottom of the page.
        :param
            -   y_position: integer - The position the browser will scroll to vertically.
            -   x_position: integer - The position the browser will scroll to horizontally.
            -   scroll_top: boolean - Whether or not the element will be scrolled to the top.
            -   scroll_bottom:  boolean - Whether or not the element will be scrolled to the bottom.
        """
        try:
            if scroll_top:
                self.execute_script("window.scrollTo(0, 0);")
            elif scroll_bottom:
                total_height = self.execute_script("var height = document.body.scrollHeight; return height")
                self.execute_script("window.scrollTo(0, arguments[0]);", total_height)
            elif type(y_position) == int or type(x_position) == int:
                self.execute_script("window.scrollTo(arguments[0], arguments[1]);", x_position, y_position)
            else:
                message = "Unable to scroll to position ('{0}', '{1}') on page '{2}'.".format(x_position, y_position,
                                                                                              self.driver.current_url)
                raise ScrollPositionError(msg=message, stacktrace=traceback.format_exc(),
                                          current_url=self.driver.current_url, y_position=y_position,
                                          x_position=x_position)
        except Exception as scroll_window_to_position_error:
            message = "Unable to scroll to position ('{0}', '{1}').\n" \
                      "<{2}>".format(x_position, y_position, scroll_window_to_position_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def get_window_current_scroll_position(self, get_both_positions=False, get_only_x_position=False):
        """
        This will get the current scroll position that the window is at. You can get both the X scroll position and Y
        scroll position, only the X scroll position, or only the Y scroll position.
        :param
            -   get_both_positions: boolean - Whether or not both the X position and Y position are returned.
            -   get_only_x_position: boolean - Whether or not only the X position is returned.
        :return
            -   x_scroll_position:  integer - The amount that the window has been scrolled on the x axis.
            -   y_scroll_position:  integer - The amount that the window has been scrolled on the y axis.
        """
        try:
            x_scroll_position = self.execute_script("return window.scrollX;")
            y_scroll_position = self.execute_script("return window.scrollY;")
            if get_both_positions and not get_only_x_position:
                return x_scroll_position, y_scroll_position
            elif get_only_x_position and not get_both_positions:
                return x_scroll_position
            else:
                return y_scroll_position
        except Exception as get_window_current_scroll_position_error:
            message = "Unable to determine the scroll position of the window.\n" \
                      "<{0}>".format(get_window_current_scroll_position_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def scroll_an_element(self, css_selector=None, web_element=None, y_position=0, x_position=0, scroll_padding=0,
                          scroll_top=False, scroll_bottom=False, scroll_left=False, scroll_right=False,
                          scroll_horizontal=False):
        """
        This will scroll an element on a page (e.g. An ISI modal).  The user can have it scroll to the top of the
        element, the bottom of the element, a specific position in the element, or by the height of the scrollable area
        of the element.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
            -   y_position:    integer - The position that the element will be scrolled to vertically.
            -   x_position:    integer - The position that the element will be scrolled to horizontally.
            -   scroll_padding: integer - The amount of padding that will be used when scroll by the element's height.
            -   scroll_top: boolean - Whether or not the element will be scrolled to the top.
            -   scroll_bottom:  boolean - Whether or not the element will be scrolled to the bottom.
            -   scroll_left:  boolean - Whether or not the element will be scrolled to the left.
            -   scroll_right:  boolean - Whether or not the element will be scrolled to the right.
            -   scroll_horizontal:  boolean - Whether or not the element will be scrolled to the horizontally.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.ensure_element_visible(web_element=web_element, css_selector=css_selector)
            if scroll_top:
                self.execute_script("arguments[0].scrollTop = 0;", web_element)
            elif scroll_bottom:
                element_max_height = self.execute_script("var element = arguments[0]; "
                                                         "var scrollHeight = element.scrollHeight; "
                                                         "var clientHeight = element.clientHeight; "
                                                         "var maxHeight = scrollHeight - clientHeight; "
                                                         "return maxHeight;", web_element)
                self.execute_script("arguments[0].scrollTop = arguments[1];", web_element, element_max_height)
            elif scroll_left:
                self.execute_script("arguments[0].scrollLeft = 0;", web_element)
            elif scroll_right:
                element_max_width = self.execute_script("var element = arguments[0]; "
                                                        "var scrollWidth = element.scrollWidth; "
                                                        "var clientWidth = element.clientWidth; "
                                                        "var maxWidth = scrollWidth - clientWidth; "
                                                        "return maxWidth;", web_element)
                self.execute_script("arguments[0].scrollLeft = arguments[1];", web_element, element_max_width)
            elif y_position or x_position:
                self.execute_script("arguments[0].scrollTop = arguments[1];", web_element, y_position)
                self.execute_script("arguments[0].scrollLeft = arguments[1];", web_element, x_position)
            elif scroll_horizontal:
                element_width = self.execute_script("var element = arguments[0]; "
                                                    "var elementWidth = element.offsetWidth; "
                                                    "return elementWidth;", web_element)
                self.execute_script("arguments[0].scrollLeft += (arguments[1] - arguments[2]);",
                                    web_element, element_width, scroll_padding)
            else:
                element_height = self.execute_script("var element = arguments[0]; "
                                                     "var elementHeight = element.offsetHeight; "
                                                     "return elementHeight;", web_element)
                self.execute_script("arguments[0].scrollTop += (arguments[1] - arguments[2]);",
                                    web_element, element_height, scroll_padding)
        except SeleniumHelperExceptions as scroll_element_error:
            scroll_element_error.msg = "Unable to scroll element. | " + scroll_element_error.msg
            raise scroll_element_error
        except Exception as unexpected_error:
            message = "Unable to scroll the element on page '{0}'.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def get_element_current_scroll_position(self, css_selector=None, web_element=None, get_both_positions=False,
                                            get_only_x_position=False):
        """
        Check to see what position the scrollable element is at.You can get both the X scroll position and Y
        scroll position, only the X scroll position, or only the Y scroll position.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
            -   get_both_positions: boolean - Whether or not both the X position and Y position are returned.
            -   get_only_x_position:    boolean - Whether or not only the X position is returned.
        :return
            -   x_scroll_position:  integer - The amount that the element has been scrolled on the x axis.
            -   y_scroll_position:  integer - The amount that the element has been scrolled on the y axis.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.ensure_element_visible(web_element=web_element, css_selector=css_selector)
            x_scroll_position = self.execute_script("var element = arguments[0]; "
                                                    "scrollPosition = element.scrollLeft; "
                                                    "return scrollPosition;", web_element)
            y_scroll_position = self.execute_script("var element = arguments[0]; "
                                                    "scrollPosition = element.scrollTop; "
                                                    "return scrollPosition;", web_element)
            if get_both_positions and not get_only_x_position:
                return x_scroll_position, y_scroll_position
            elif get_only_x_position and not get_both_positions:
                return x_scroll_position
            else:
                return y_scroll_position
        except SeleniumHelperExceptions as current_scroll_error:
            current_scroll_error.msg = "Unable to determine the element's scroll position. | " + \
                                       current_scroll_error.msg
            raise current_scroll_error
        except Exception as unexpected_error:
            message = "Unable to determine the scroll position of the element on page '{0}'.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def get_is_element_scroll_position_at_top(self, css_selector=None, web_element=None):
        """
        Check to see if the scroll position is at the top of the scrollable element.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
        :return
            -   at_top: boolean - Whether or not the scrollable element is at the top.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.ensure_element_visible(web_element=web_element, css_selector=css_selector)
            scroll_position = self.execute_script("var element = arguments[0]; "
                                                  "scrollPosition = element.scrollTop; "
                                                  "return scrollPosition;", web_element)
            if scroll_position != 0:
                return False
            else:
                return True
        except SeleniumHelperExceptions as scroll_at_top_error:
            scroll_at_top_error.msg = "Unable to determine if element is scrolled to the top. | " + \
                                      scroll_at_top_error.msg
            raise scroll_at_top_error
        except Exception as unexpected_error:
            message = "Unable to determine if the scroll position of the element on page '{0}' is at the top.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def get_is_element_scroll_position_at_bottom(self, css_selector=None, web_element=None):
        """
        Check to see if the scroll position is at the bottom of the scrollable element.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
        :return
            -   at_bottom:  boolean - Whether or not the scrollable element is at the bottom.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.ensure_element_visible(web_element=web_element, css_selector=css_selector)
            element_max_height = self.execute_script("var element = arguments[0]; "
                                                     "var scrollHeight = element.scrollHeight; "
                                                     "var clientHeight = element.clientHeight; "
                                                     "var maxHeight = scrollHeight - clientHeight; "
                                                     "return maxHeight;", web_element)
            scroll_position = self.execute_script("var element = arguments[0]; "
                                                  "var scrollPosition = element.scrollTop; "
                                                  "return scrollPosition;", web_element)
            if scroll_position < element_max_height - 1:
                return False
            else:
                return True
        except SeleniumHelperExceptions as scroll_at_bottom_error:
            scroll_at_bottom_error.msg = "Unable to determine if element is scrolled to the bottom. | " + \
                                         scroll_at_bottom_error.msg
            raise scroll_at_bottom_error
        except Exception as unexpected_error:
            message = "Unable to determine if the scroll position of the element on page '{0}' is at the bottom."\
                      "\n<{1}>".format(self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def get_is_element_scroll_position_at_most_right(self, css_selector=None, web_element=None):
        """
        Check to see if the scroll position is at the most right of the scrollable element.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
        :return
            -   at_right:  boolean - Whether or not the scrollable element is at the most right.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.ensure_element_visible(web_element=web_element, css_selector=css_selector)
            element_max_width = self.execute_script("var element = arguments[0]; "
                                                    "var scrollWidth = element.scrollWidth; "
                                                    "var clientWidth = element.clientWidth; "
                                                    "var maxWidth = scrollWidth - clientWidth; "
                                                    "return maxWidth;", web_element)
            scroll_position = self.execute_script("var element = arguments[0]; "
                                                  "var scrollPosition = element.scrollLeft; "
                                                  "return scrollPosition;", web_element)
            if scroll_position < element_max_width - 1:
                return False
            else:
                return True
        except SeleniumHelperExceptions as scroll_at_right_error:
            scroll_at_right_error.msg = "Unable to determine if element is scrolled to the most right. | " + \
                                         scroll_at_right_error.msg
            raise scroll_at_right_error
        except Exception as unexpected_error:
            message = "Unable to determine if the scroll position of the element on page '{0}' is at the most right." \
                      "\n<{1}>".format(self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def hide_element(self, css_selector=None, web_element=None):
        """
        This will hide a specified element.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.ensure_element_visible(web_element=web_element, css_selector=css_selector)
            self.execute_script("arguments[0].style.display = 'none';", web_element)
        except SeleniumHelperExceptions as hide_error:
            hide_error.msg = "Unable to hide element. | " + hide_error.msg
            raise hide_error
        except Exception as unexpected_error:
            message = "Unable to hide element on page '{0}', it may already hidden.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)
            message += " | Based off the CSS Selector: '{0}' or WebElement passed through.".format(css_selector)
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def show_element(self, css_selector=None, web_element=None):
        """
        This will show a specified element.
        :param
            -   css_selector:   string - The specific element that will be interacted with.
            -   web_element:    object - The WebElement that will be interacted with.
        """
        try:
            if css_selector and not web_element:
                web_element = self.get_element(css_selector)
            self.execute_script("arguments[0].style.display = 'block';", web_element)
        except SeleniumHelperExceptions as show_error:
            show_error.msg = "Unable to show element. | " + show_error.msg
            raise show_error
        except Exception as unexpected_error:
            message = "Unable to show element on page '{0}', that element may not exist.\n" \
                      "<{1}>".format(self.driver.current_url, unexpected_error)
            if css_selector:
                message += " | CSS Selector: '{0}'".format(css_selector)
            else:
                message += " | Based off the WebElement passed through."
            raise ElementError(msg=message, stacktrace=traceback.format_exc(),
                               current_url=self.driver.current_url, css_selector=css_selector)

    def add_cookie(self, name=None, value=None):
        """
            This will show a specified element.
            :param
                -   name:   string - Name of the cookie you want to pass in .
                -   value:    string - Valye of the cookie.
       """
        try:
            url = self.get_current_url()
            url_parse = urlparse.urlparse(url)
            domain = ".{}".format(url_parse.netloc.split(".", 1)[-1])
            path = url_parse.path
            self.driver.add_cookie({"name": name, "value": value, "domain": domain, "path": path})
        except Exception as cookie_error:
            message = "There was an issue creating a a cookie: {0}".format(cookie_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())

    def delete_cookie(self, name=None):
        """
            This will show a specified element.
            :param
                -   name:   string - Name of the cookie you want to delete .
       """
        try:
            self.driver.delete_cookie(name)
        except Exception as cookie_error:
            message = "There was an issue deleting a a cookie: {0}".format(cookie_error)
            raise DriverAttributeError(msg=message, stacktrace=traceback.format_exc())


class SeleniumHelperExceptions(common.exceptions.WebDriverException):
    def __init__(self, msg, stacktrace, current_url):
        self.current_url = current_url
        self.details = {"current_url": self.current_url, "stacktrace": stacktrace}
        super(SeleniumHelperExceptions, self).__init__(msg=msg, stacktrace=stacktrace)

    def __str__(self):
        exception_msg = "Selenium Exception: \n"
        detail_string = "Exception Details:\n"
        for key, value in self.details.items():
            detail_string += "{0}: {1}\n".format(key, value)
        exception_msg += detail_string
        exception_msg += "Message: {0}".format(self.msg)

        return exception_msg


class ElementError(SeleniumHelperExceptions):
    def __init__(self, msg, stacktrace, current_url, css_selector):
        super(ElementError, self).__init__(msg=msg, stacktrace=stacktrace, current_url=current_url)
        self.css_selector = css_selector
        self.details["css_selector"] = self.css_selector


class ElementNotVisibleError(SeleniumHelperExceptions):
    def __init__(self, msg, stacktrace, current_url, css_selector):
        super(ElementNotVisibleError, self).__init__(msg=msg, stacktrace=stacktrace, current_url=current_url)
        self.css_selector = css_selector
        self.details["css_selector"] = self.css_selector


class TimeoutError(SeleniumHelperExceptions):
    def __init__(self, msg, stacktrace, current_url, css_selector, wait_time):
        super(TimeoutError, self).__init__(msg=msg, stacktrace=stacktrace, current_url=current_url)
        self.css_selector = css_selector
        self.wait_time = wait_time
        self.details["css_selector"] = self.css_selector
        self.details["wait_time"] = self.wait_time


class ClickPositionError(SeleniumHelperExceptions):
    def __init__(self, msg, stacktrace, current_url, css_selector, y_position, x_position):
        super(ClickPositionError, self).__init__(msg=msg, stacktrace=stacktrace, current_url=current_url)
        self.css_selector = css_selector
        self.y_position = y_position
        self.x_position = x_position
        self.details["css_selector"] = self.css_selector
        self.details["y_position"] = self.y_position
        self.details["x_position"] = self.x_position


class CursorLocationError(SeleniumHelperExceptions):
    def __init__(self, msg, stacktrace, current_url, x_position, y_position):
        super(CursorLocationError, self).__init__(msg=msg, stacktrace=stacktrace, current_url=current_url)
        self.x_position = x_position
        self.y_position = y_position
        self.details["x_position"] = self.x_position
        self.details["y_position"] = self.y_position


class ScrollPositionError(SeleniumHelperExceptions):
    def __init__(self, msg, stacktrace, current_url, y_position, x_position):
        super(ScrollPositionError, self).__init__(msg=msg, stacktrace=stacktrace, current_url=current_url)
        self.y_position = y_position
        self.x_position = x_position
        self.details["y_position"] = self.y_position
        self.details["x_position"] = self.x_position

class ScreenshotError(SeleniumHelperExceptions):
    def __init__(self, msg, stacktrace, current_url, file_name, file_path):
        super(ScreenshotError, self).__init__(msg=msg, stacktrace=stacktrace, current_url=current_url)
        self.file_name = file_name
        self.file_path = file_path
        self.details["file_name"] = self.file_name
        self.details["file_path"] = self.file_path


class DriverExceptions(Exception):
    def __init__(self, msg, stacktrace=None, details=None):
        self.msg = msg
        self.details = {} if details is None else details
        self.stacktrace = stacktrace
        super(DriverExceptions, self).__init__()

    def __str__(self):
        exception_msg = "Driver Creation Exception: \n"
        if self.stacktrace is not None:
            exception_msg += "\nStacktrace: {0}\n".format(self.stacktrace)
        if self.details:
            detail_string = "\nException Details:\n"
            for key, value in self.details.items():
                detail_string += "{0}: {1}\n".format(key, value)
            exception_msg += detail_string
        exception_msg += "Message: {0}".format(self.msg)

        return exception_msg


class DriverAttributeError(DriverExceptions):
    def __init__(self, msg, stacktrace=None):
        super(DriverAttributeError, self).__init__(msg=msg, stacktrace=stacktrace)


class DriverSizeError(DriverExceptions):
    def __init__(self, msg, stacktrace=None, width=None, height=None):
        super(DriverSizeError, self).__init__(msg=msg, stacktrace=stacktrace)
        self.width = width
        self.height = height
        self.details["width"] = self.width
        self.details["height"] = self.height


class DriverURLError(DriverExceptions):
    def __init__(self, msg, stacktrace=None, desired_url=None):
        super(DriverURLError, self).__init__(msg=msg, stacktrace=stacktrace)
        self.desired_url = desired_url
        self.details["desired_url"] = self.desired_url



import unittest
from datetime import datetime, timedelta
from the_ark.field_handlers import CHECK_BOX_FIELD, SELECT_FIELD, STRING_FIELD, PHONE_FIELD, ZIP_CODE_FIELD, \
    DATE_FIELD, INTEGER_FIELD, EMAIL_FIELD, PASSWORD_FIELD
from the_ark import input_generator as ig
from the_ark.input_generator import InputGeneratorException
from mock import patch


class InputGeneratorTestCase(unittest.TestCase):
    @patch("random.random")
    def test_set_leave_blank(self, random_value):
        # - Tests without field data
        fake_blank = ig.set_leave_blank(1, True)
        self.assertEqual(fake_blank, False)

        # - Tests with field data
        # Not Required
        # - Blank
        random_value.return_value = 0.20
        fake_blank = ig.set_leave_blank(5, False)
        self.assertEqual(fake_blank, True)
        # - Not Blank
        random_value.return_value = 0.20
        fake_blank = ig.set_leave_blank(1, False)
        self.assertEqual(fake_blank, False)
        # -- Required
        fake_blank = ig.set_leave_blank(5, True)
        self.assertEqual(fake_blank, False)

    def test_check_min_vs_max(self):
        # - Test without a field
        with self.assertRaises(ig.InputGeneratorException):
            ig.check_min_vs_max(13, 12)

    # ===================================================================
    # --- dispatch_field() Tests
    # ===================================================================
    # - 'required' variable Test
    @patch("the_ark.input_generator.generate_string")
    def test_dispatch_string_field_required(self, string_method):
        field_data = {"type": STRING_FIELD,
                      "min": 10,
                      "max": 15,
                      "required": True}
        string_method.return_value = "return_value"
        ig.dispatch_field(field_data, 1)
        string_method.assert_called_once_with(field_data["min"], field_data["max"], 1, True)

    @patch("the_ark.input_generator.generate_string")
    def test_dispatch_string_field(self, string_method):
        field_data = {"type": STRING_FIELD,
                      "min": 10,
                      "max": 15}
        string_method.return_value = "return_value"
        ig.dispatch_field(field_data, 1)
        string_method.assert_called_once_with(field_data["min"], field_data["max"], 1, False)

    # - Dispatch Integer
    @patch("the_ark.input_generator.generate_integer")
    def test_dispatch_integer_field_without_padding(self, integer_method):
        field_data = {"type": INTEGER_FIELD,
                      "min": 10,
                      "max": 15}

        integer_method.return_value = 3
        ig.dispatch_field(field_data, 1)
        integer_method.assert_called_once_with(field_data["min"], field_data["max"], 1, 1, False)

    @patch("the_ark.input_generator.generate_integer")
    def test_dispatch_integer_field_with_padding(self, integer_method):
        field_data = {"type": INTEGER_FIELD,
                      "min": 10,
                      "max": 15,
                      "padding": 10}

        integer_method.return_value = 3
        ig.dispatch_field(field_data, 1)
        integer_method.assert_called_once_with(field_data["min"], field_data["max"], 10, 1, False)

    # - Dispatch E-mail
    @patch("the_ark.input_generator.generate_email")
    def test_dispatch_email_field_without_domain(self, email_method):
        field_data = {"type": EMAIL_FIELD}
        email_method.return_value = "test@{0}".format(ig.DEFAULT_DOMAIN)
        ig.dispatch_field(field_data, 1)
        email_method.assert_called_once_with(ig.DEFAULT_DOMAIN, 1, False)

    @patch("the_ark.input_generator.generate_email")
    def test_dispatch_email_field_with_custom_domain(self, email_method):
        field_data = {"type": EMAIL_FIELD,
                      "domain": "gmail.com"}
        email_method.return_value = "test@gmail.com"
        ig.dispatch_field(field_data, 1)
        email_method.assert_called_once_with(field_data["domain"], 1, False)

    # - Dispatch Password
    def test_dispatch_password_generator_length(self):
        field_data = {"type": PASSWORD_FIELD}
        returned_password = ig.dispatch_field(field_data)
        self.assertTrue(len(returned_password) >= 6)

    def test_dispatch_password_generator_special_character(self):
        field_data = {"type": PASSWORD_FIELD}
        returned_password = ig.dispatch_field(field_data)
        self.assertTrue(any(character in returned_password for character in ig.SPECIAL_CHARACTER_LIST))

    @patch("the_ark.input_generator.generate_string")
    def test_dispatch_password_generator_error(self, string_method):
        string_method.side_effect = Exception('Boom!')
        with self.assertRaises(ig.InputGeneratorException):
            ig.generate_password()

    # - Dispatch Phone Number
    @patch("the_ark.input_generator.generate_phone")
    def test_dispatch_phone_field_defaults(self, phone_method):
        field_data = {"type": PHONE_FIELD}
        phone_method.return_value = "9878767654"
        ig.dispatch_field(field_data, 1)
        phone_method.assert_called_once_with(False, False, False, False, 1, False)

    @patch("the_ark.input_generator.generate_phone")
    def test_dispatch_phone_field_override_variables(self, phone_method):
        # - Using int values here instead of bools to verify determine the right params are changed
        field_data = {"type": PHONE_FIELD,
                      "decimal": 1,
                      "parenthesis": 2,
                      "dash": 3,
                      "space": 4}
        phone_method.return_value = "9878767654"
        ig.dispatch_field(field_data, 1)
        phone_method.assert_called_once_with(field_data["decimal"],
                                             field_data["parenthesis"],
                                             field_data["dash"],
                                             field_data["space"], 1, False)

    # - Dispatch ZIP Code
    @patch("the_ark.input_generator.generate_zip_code")
    def test_dispatch_zip_code_field_defaults(self, zip_method):
        field_data = {"type": ZIP_CODE_FIELD}
        zip_method.return_value = "99999"
        ig.dispatch_field(field_data, 1)
        zip_method.assert_called_once_with(1, False)

    # - Dispatch Index Field
    @patch("the_ark.input_generator.generate_index")
    def test_dispatch_index_field_without_random(self, index_method):
        field_data = {"type": SELECT_FIELD,
                      "enum": ["selector1", "selector2"]}
        index_method.return_value = 1
        ig.dispatch_field(field_data, 20)
        index_method.assert_called_once_with(len(field_data["enum"]), 20, False, False)

    @patch("the_ark.input_generator.generate_index")
    def test_dispatch_index_field_with_random(self, index_method):
        field_data = {"type": SELECT_FIELD,
                      "enum": ["selector1", "selector2"],
                      "random": True}
        index_method.return_value = 1
        ig.dispatch_field(field_data, 20)
        index_method.assert_called_once_with(len(field_data["enum"]), 20, False, True)

    # - Dispatch Check Box
    @patch("the_ark.input_generator.generate_check_box")
    def test_dispatch_check_box_field(self, check_box_method):
        field_data = {"type": CHECK_BOX_FIELD,
                      "enum": [{"selector1": "Selector 1"},
                               {"selector2": "Selector 2"}]}
        check_box_method.return_value = 1
        ig.dispatch_field(field_data, 150)
        check_box_method.assert_called_once_with(len(field_data["enum"]), 150, False)

    # - Dispatch Date
    @patch("the_ark.input_generator.generate_date")
    def test_dispatch_date_field_defaults(self, date_method):
        field_data = {"type": DATE_FIELD}
        date_method.return_value = "11/19/1911"
        ig.dispatch_field(field_data, 1)
        date_method.assert_called_once_with(ig.DEFAULT_START_DATE,
                                            ig.DEFAULT_END_DATE,
                                            ig.DEFAULT_DATE_FORMAT, 1, False)

    @patch("the_ark.input_generator.generate_date")
    def test_dispatch_date_field_override_variables(self, date_method):
        # - Using int values here instead of bools to verify determine the right params are changed
        field_data = {"type": DATE_FIELD,
                      "start_date": 1,
                      "end_date": 2,
                      "date_format": 3}
        date_method.return_value = "11/19/1911"
        ig.dispatch_field(field_data, 1)
        date_method.assert_called_once_with(field_data["start_date"],
                                            field_data["end_date"],
                                            field_data["date_format"], 1, False)

    # - Exceptions
    def test_dispatch_invalid_type(self):
        field_data = {"type": "pickles"}
        with self.assertRaises(ig.UnknownFieldType):
            ig.dispatch_field(field_data)

    @patch("the_ark.input_generator.generate_string")
    def test_dispatch_field_handler_exception_without_name(self, string_method):
        field_data = {"type": STRING_FIELD,
                      "min": 10,
                      "max": 12}
        string_method.side_effect = ig.MissingKey("Hello", "key")
        with self.assertRaises(ig.InputGeneratorException):
            ig.dispatch_field(field_data)

    @patch("the_ark.input_generator.generate_string")
    def test_dispatch_field_handler_exception_with_name(self, string_method):
        field_data = {"type": STRING_FIELD,
                      "name": "First Name",
                      "min": 10,
                      "max": 12}
        string_method.side_effect = ig.MissingKey("Hello", "key")
        with self.assertRaises(ig.InputGeneratorException):
            ig.dispatch_field(field_data)

    def test_dispatch_key_error_without_name(self):
        field_data = {"type": STRING_FIELD,
                      "max": 12}

        with self.assertRaises(ig.InputGeneratorException):
            ig.dispatch_field(field_data)

    def test_dispatch_key_error_with_name(self):
        field_data = {"type": STRING_FIELD,
                      "name": "First Name",
                      "max": 12}

        with self.assertRaises(ig.InputGeneratorException):
            ig.dispatch_field(field_data)

    @patch("the_ark.input_generator.generate_string")
    def test_dispatch_general_exception_without_name(self, string_method):
        field_data = {"type": STRING_FIELD,
                      "min": 10,
                      "max": 12}

        string_method.side_effect = Exception("Boo!")
        with self.assertRaises(ig.InputGeneratorException):
            ig.dispatch_field(field_data)

    @patch("the_ark.input_generator.generate_string")
    def test_dispatch_general_exception_exception_with_name(self, string_method):
        field_data = {"type": STRING_FIELD,
                      "name": "First Name",
                      "min": 10,
                      "max": 12}

        string_method.side_effect = Exception("Boo!")
        with self.assertRaises(ig.InputGeneratorException):
            ig.dispatch_field(field_data)

    # ===================================================================
    # --- Generate Input Tests
    # ===================================================================
    # - Generate String
    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_string_default_values(self, leave_blank):
        leave_blank.return_value = False
        returned_string = ig.generate_string()
        if not ig.DEFAULT_STRING_MIN <= len(returned_string) <= ig.DEFAULT_STRING_MAX:
            self.fail("The string length returned using the method defaults was incorrect. "
                      "It should be between {0} and {1} but was {2}".format(ig.DEFAULT_STRING_MIN,
                                                                            ig.DEFAULT_STRING_MAX,
                                                                            len(returned_string)))

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_string_blank_return(self, leave_blank):
        leave_blank.return_value = True
        self.assertEqual("", ig.generate_string())

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_string_specific_length(self, leave_blank):
        leave_blank.return_value = False
        # - Test string length ranges
        returned_string = ig.generate_string(15, 15)
        self.assertEqual(len(returned_string), 15)

    @patch("the_ark.input_generator.check_min_vs_max")
    def test_generate_string_general_exception(self, min_max):
        min_max.side_effect = Exception("Boo!")
        with self.assertRaises(ig.InputGeneratorException):
            ig.generate_string()

    # -- Generate Integer
    def test_generate_integer(self):
        # - Test default values
        returned_integer = ig.generate_integer()
        if not ig.DEFAULT_INTEGER_MIN <= len(returned_integer) <= ig.DEFAULT_INTEGER_MAX:
            self.fail("The integer length returned using the method defaults was incorrect. "
                      "It should be between {0} and {1} but was {2}".format(ig.DEFAULT_INTEGER_MIN,
                                                                            ig.DEFAULT_INTEGER_MAX,
                                                                            len(returned_integer)))

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_integer_blank_return(self, leave_blank):
        leave_blank.return_value = True
        self.assertEqual("", ig.generate_integer())

    def test_generate_integer_specific_range(self):
        returned_integer = ig.generate_integer(15, 15)
        self.assertEqual(returned_integer, "15")

    def test_generate_integer_with_padding(self):
        padding = 4
        returned_integer = ig.generate_integer(3, 8, padding)
        self.assertEqual(len(returned_integer), padding)

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_integer_exception(self, leave_blank):
        leave_blank.side_effect = Exception("Boo!")
        with self.assertRaises(ig.InputGeneratorException):
            ig.generate_integer("apples", "oranges")

    # - Generate Email
    def test_generate_email_default_values(self):
        returned_email = ig.generate_email()
        if ig.DEFAULT_DOMAIN not in returned_email:
            self.assertIn(ig.DEFAULT_DOMAIN, returned_email,
                          "The generated email did not include the default domain when the default parameters were "
                          "used. It should contain '{0}' but the returned email was: '{1}'".format(ig.DEFAULT_DOMAIN,
                                                                                                   returned_email))

    def test_generate_email_custom_domain(self):
        test_domain = "vincentrocks.com"
        returned_email = ig.generate_email(test_domain)
        if ig.DEFAULT_DOMAIN not in returned_email:
            self.assertIn(test_domain, returned_email,
                          "The generated email did not include the custom domain when a domain was passed in."
                          "It should contain '{0}' but the returned email was: '{1}'".format(test_domain,
                                                                                             returned_email))

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_email_blank_return(self, leave_blank):
        leave_blank.return_value = True
        returned_email = ig.generate_email()
        self.assertEqual("", returned_email)

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_email_exception(self, leave_blank):
        leave_blank.side_effect = Exception('Boom!')
        self.assertRaises(ig.InputGeneratorException, ig.generate_email)

    # -- Generate Phone
    def test_generate_phone_default_values(self):
        # Test default values ##########
        returned_phone = ig.generate_phone()
        self.assertRegexpMatches(returned_phone, "^\d{10}")

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_phone_blank_return(self, leave_blank):
        leave_blank.return_value = True
        returned_phone = ig.generate_phone()
        self.assertEqual(returned_phone, "")

    def test_generate_phone_all_dashed(self):
        # Check for ### -### -####
        returned_phone = ig.generate_phone(dash=True)
        self.assertRegexpMatches(returned_phone, "^[2-9]\d{2}-\d{3}-\d{4}")

    def test_generate_phone_all_spaces(self):
        # Check for ### ### ####
        returned_phone = ig.generate_phone(space=True)
        self.assertRegexpMatches(returned_phone, "^[2-9]\d{2}\s\d{3}\s\d{4}")

    def test_generate_phone_parenthesis_and_dash(self):
        # Check for (###)### -####
        returned_phone = ig.generate_phone(parenthesis=True, dash=True)
        self.assertRegexpMatches(returned_phone, "^\(\d{3}\)\d{3}-\d{4}")

    def test_generate_phone_just_parenthesis(self):
        # Check for (###)#######
        returned_phone = ig.generate_phone(parenthesis=True)
        self.assertRegexpMatches(returned_phone, "^\(\d{3}\)\d{7}")

    def test_generate_phone_parenthesis_dash_space(self):
        # Check for (###) ### -####
        returned_phone = ig.generate_phone(parenthesis=True, dash=True, space=True)
        self.assertRegexpMatches(returned_phone, "\(\d{3}\)\s\d{3}-\d{4}")

    def test_generate_phone_all_decimals(self):
        # Check for ###.###.####
        returned_phone = ig.generate_phone(decimals=True)
        self.assertRegexpMatches(returned_phone, "^[2-9]\d{2}\.\d{3}\.\d{4}")

    def test_generate_phone_decimal_override(self):
        # Decimals override everything
        returned_phone = ig.generate_phone(decimals=True, parenthesis=True, dash=True, space=True)
        self.assertRegexpMatches(returned_phone, "^[2-9]\d{2}\.\d{3}\.\d{4}")

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_phone_exception(self, leave_blank):
        leave_blank.side_effect = Exception('Boom!')
        with self.assertRaises(ig.InputGeneratorException):
            ig.generate_phone(1)

    # - Generate ZIP Code
    def test_generate_zip_default_values(self):
        returned_zip = ig.generate_zip_code()
        self.assertEqual(len(returned_zip), 5)

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_zip_blank_return(self, leave_blank):
        leave_blank.return_value = True
        returned_zip = ig.generate_zip_code()
        self.assertEqual("", returned_zip)

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_zip_exception(self, leave_blank):
        # Test the General Exception catch
        leave_blank.side_effect = Exception('Boom!')
        with self.assertRaises(ig.InputGeneratorException):
            ig.generate_zip_code()

    # - Generate Index
    def test_generate_index_default_values(self):
        returned_index = ig.generate_index()
        if returned_index not in range(0, 2):
            self.fail("The index generated using the method defaults was incorrect. "
                      "It should be between 0 and {0} but was {1}".format(ig.DEFAULT_INDEX_OPTIONS, returned_index))

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_index_blank_return(self, leave_blank):
        leave_blank.return_value = True
        self.assertEqual("", ig.generate_index())

    def test_generate_index_last_option(self):
        index = 5
        self.assertEqual(ig.generate_index(index, 5), index - 1)

    def test_generate_index_always_random(self):
        generated_indexes = []
        for i in range(0, 10):
            generated_indexes.append(ig.generate_index(100, always_random=True))
        if all(x == generated_indexes[0] for x in generated_indexes):
            self.fail("The indexes generated in the Always Random check were not different!")

    def test_generate_index_all_options_used(self):
        generated_indexes = []
        for i in range(0, 10):
            generated_indexes.append(ig.generate_index(100, test_number=1000))
        if all(x == generated_indexes[0] for x in generated_indexes):
            self.fail("The indexes generated in the Always Random check were not different!")

    def test_generate_index_exception(self):
        with self.assertRaises(ig.InputGeneratorException):
            ig.generate_index(["apples", "pickles"], 500)

    # - Generate Check Box
    def test_generate_check_box_default_values(self):
        returned_index = ig.generate_check_box()
        if not isinstance(returned_index, list):
            self.fail("The index array generated by the default parameters was not... an array. "
                      "It should be a list, but this was returned: {0}".format(returned_index))

    @patch("random.random")
    def test_generate_check_box_not_required_blank(self, random_result):
        random_result.return_value = .20
        self.assertEqual([], ig.generate_check_box(2, 10, False))

    @patch("random.random")
    def test_generate_check_box_not_required_not_blank(self, random_result):
        random_result.return_value = .50
        self.assertIsInstance(ig.generate_check_box(2, 10, False), list)

    @patch("random.random")
    def test_generate_check_box_random_choice_select_each(self, random_result):
        random_result.return_value = .50
        generated_indexes = ig.generate_check_box(5, 10)
        self.assertEqual(len(generated_indexes), 5)

    @patch("random.random")
    def test_generate_check_box_random_choice_select_all(self, random_result):
        random_result.return_value = .50
        self.assertEqual([1], ig.generate_check_box(2, 2))

    def test_generate_check_box_exception(self):
        with self.assertRaises(ig.InputGeneratorException):
            ig.generate_check_box("apples")

    # - Generate Date
    def test_generate_date(self):
        returned_date = ig.generate_date()
        try:
            datetime.strptime(returned_date, ig.DEFAULT_DATE_FORMAT)
        except Exception as e:
            self.fail("The default settings for generate_date() did not return a date in the default format "
                      "of {0}. The date returned was '{1}': {2}".format(ig.DEFAULT_DATE_FORMAT, returned_date, e))

    @patch("the_ark.input_generator.set_leave_blank")
    def test_generate_date_blank_return(self, leave_blank):
        leave_blank.return_value = True
        returned_date = ig.generate_date()
        self.assertEqual("", returned_date, "The check for blank dates did not return a blank string")

    def test_generate_date_range_past_int(self):
        start_date = 4
        end_date = 2
        expected_date_format = ig.DEFAULT_DATE_FORMAT
        returned_date = ig.generate_date(start_date, end_date)
        try:
            formatted_date = datetime.strptime(returned_date, ig.DEFAULT_DATE_FORMAT).date()
        except Exception as e:
            self.fail("When checking the date range, could not convert the returned date into '{0}' format in "
                      "order to compare the dates. The returned date was {1}: {2}".format(expected_date_format,
                                                                                          returned_date, e))
        # Returned Date older than current date
        self.assertGreater(datetime.now().date(), formatted_date,
                           "When using days, in the past, the returned date was not older than the current date")
        # Returned Date older than end date
        self.assertGreater((datetime.now() - timedelta(days=end_date)).date(),
                           formatted_date,
                           "When using days, in the past, the returned date was not older than the end date")

    def test_generate_date_range_past_date(self):
        # - Past in dates (%Y-%m-%d) with "%Y/%d/%m" format
        expected_format = "%Y/%d/%m"
        start_date = datetime.strptime("19800404", "%Y%m%d").date()
        end_date = datetime.strptime("20150325", "%Y%m%d").date()
        returned_date = ig.generate_date(str(start_date), str(end_date), expected_format)
        # - Check returned date format
        try:
            formatted_date = datetime.strptime(returned_date, expected_format).date()
        except Exception as e:
            self.fail("When checking the date range with custom format, could not convert the returned date into '{0}'"
                      " format in order to compare the dates. The returned date was {1}: {2}"
                      .format(expected_format, returned_date, e))
        # Returned Date older than current date
        self.assertGreater(formatted_date, start_date,
                           "When using dates, in the past, the returned date was not older than the start date")
        # Returned Date older than end date
        self.assertGreater(end_date, formatted_date,
                           "When using dates, in the past, the returned date was not older than the end date")

    def test_generate_date_range_future_int(self):
        # - Date Range Checks (in the FUTURE!!! *cue sci-fi sound effects)
        start_date = -4
        end_date = -20
        expected_date_format = ig.DEFAULT_DATE_FORMAT
        returned_date = ig.generate_date(start_date, end_date)
        try:
            formatted_date = datetime.strptime(returned_date, ig.DEFAULT_DATE_FORMAT).date()
        except Exception as e:
            self.fail("When checking the date range in the future, could not convert the returned date into '{0}' "
                      "format in order to compare the dates. The returned date was {1}: {2}"
                      .format(expected_date_format, returned_date, e))
        # Returned Date later than current date
        self.assertGreater(formatted_date, datetime.now().date(),
                           "When using days, in the future, the returned date was not later than the current date")
        # Returned Date older than end date
        self.assertGreater((datetime.now() - timedelta(days=end_date)).date(), formatted_date,
                           "When using days, in the past, the returned date was not older than the end date")

    def test_generate_date_range_future_date(self):
        # Future in dates (%Y-%m-%d) with "%m-%d-%y" format
        expected_future_format = "%m-%d-%y"
        start_date = datetime.strptime("20250404", "%Y%m%d").date()
        end_date = datetime.strptime("20300325", "%Y%m%d").date()
        returned_date = ig.generate_date(str(start_date), str(end_date), expected_future_format)
        # - Check returned date format
        try:
            formatted_date = datetime.strptime(returned_date, expected_future_format).date()
        except Exception as e:
            self.fail("When checking the future date range with custom format, could not convert the returned date "
                      "into '{0}' format in order to compare the dates. The returned date was {1}: {2}"
                      .format(expected_future_format, returned_date, e))
        # Returned Date later than current date
        self.assertGreater(formatted_date, datetime.now().date(),
                           "When using dates, in the future, the returned date was not later than the start date")
        # Returned Date older than end date
        self.assertGreater(end_date, formatted_date,
                           "When using dates, in the future, the returned date was not older than the end date")

    def test_generate_date_date_exception(self):
        # - Without field
        with self.assertRaises(ig.InputGeneratorException):
            ig.generate_date("apples")

    # ===================================================================
    # --- Field Handler Exception
    # ===================================================================
    def test_input_generator_exception_to_string_without_details(self):
        input_exc = InputGeneratorException("Message text")
        error_string = input_exc.__str__()
        self.assertNotIn("stacktrace", error_string)

    def test_field_handler_exception_to_string_with_details(self):
        input_exc = InputGeneratorException("message",
                                            "stacktrace:\nLine 1\nLine 2",
                                            {"css_selector": "selector.1"})
        error_string = input_exc.__str__()
        self.assertIn("css_selector", error_string)
        self.assertIn("stacktrace", error_string)



import mandrill
import re
import traceback

EMAIL_REGEX = r"^[a-zA-Z0-9!#$%&\'*+/=?^_`{|}~-]+(?:\.[a-zA-Z0-9!#$%&\'*+/=?^_`{|}~-]+)*" \
              r"@(?:[a-zA-Z0-9](?:[a-zA-Z0-9-]*[a-zA-Z0-9])?\.)+[a-zA-Z0-9](?:[a-zA-Z0-9-]*[a-zA-Z0-9])?$"


class EmailClient(object):
    """
    Class that posts messages a Mandrill account
    """
    def __init__(self, api_key):
        """
        Setting up the Email connection with the Mandrill token in the configuration file.
        """
        self.mandrill_client = mandrill.Mandrill(api_key)

    def send_email(self, from_email, to_emails, message, subject="A Message From The Ark", sender_name="The Ark"):
        email_details = {"subject": subject,
                         "from_name": sender_name,
                         "from_email": from_email,
                         "to": [{"email": address} for address in to_emails],
                         "html": message}

        # Check parameters for validity
        error_message = "Email Configuration problem occurred while sending your email | "
        if not re.match(EMAIL_REGEX, from_email):
            raise EmailClientException("{0}The 'FROM' email '{1}' is not a valid email address"
                                       .format(error_message, from_email),
                                       traceback.format_exc(), email_details)
        if type(to_emails) is not list:
            raise EmailClientException("{0}The to_emails parameter is not a List object".format(error_message),
                                       traceback.format_exc(), email_details)
        for email in to_emails:
            if not re.match(EMAIL_REGEX, email):
                raise EmailClientException("{0}The 'TO' email '{1}' is not a valid email address."
                                           .format(error_message, email),
                                           traceback.format_exc(),
                                           email_details)

        # Attempt to send the email
        try:
            self.mandrill_client.messages.send(message=email_details)
        except mandrill.Error as mandrill_exception:
            message = "Mandrill Error occurred while attempting to send your email | {0}".format(mandrill_exception)
            raise EmailClientException(message, traceback.format_exc(), email_details)


class EmailClientException(Exception):
    def __init__(self, msg, stacktrace=None, details=None):
        self.msg = msg
        self.details = {} if details is None else details
        self.stacktrace = stacktrace
        super(EmailClientException, self).__init__()

    def __str__(self):
        exception_msg = "Email Client Exception: \n"
        if self.stacktrace is not None:
            exception_msg += "{0}".format(self.stacktrace)
        if self.details:
            detail_string = "\nException Details:\n"
            for key, value in self.details.items():
                detail_string += "{0}: {1}\n".format(key, value)
            exception_msg += detail_string
        exception_msg += "Message: {0}".format(self.msg)

        return exception_msg